# -*- coding: utf-8 -*-
"""
Hashing convinience functions

You should opt to use a hash*27 function over a hash* function.


TODO: the same hashing algorithm should be used everywhere
Currently there is a mix of sha1, sha256, and sha512 in different places.
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import hashlib
import copy
import os
import six
import uuid
import random
import warnings
from six.moves import zip, map
from utool import util_inject
from utool import util_path
from utool import util_type
(print, rrr, profile) = util_inject.inject2(__name__, '[hash]')

if util_type.HAVE_NUMPY:
    import numpy as np

# default length of hash codes
HASH_LEN = 16
HASH_LEN2 = 32

# HEX alphabet
ALPHABET_16 = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
               'a', 'b', 'c', 'd', 'e', 'f']

# A large base-54 alphabet (all chars are valid for filenames but not pretty)
ALPHABET_54 = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
               'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
               'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
               'u', 'v', 'w', 'x', 'y', 'z', ';', '=', '@', '[',
               ']', '^', '_', '`', '{', '}', '~', '!', '#', '$',
               '%', '&', '+', ',']


# A large base-41 alphabet (prettier subset of base 54)
ALPHABET_41 = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
               'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j',
               'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't',
               'u', 'v', 'w', 'x', 'y', 'z', '@', '!', '%', '&',
               '+']

ALPHABET_27 = [
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o',
    'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z']


# ALPHABET = ALPHABET_41
ALPHABET = ALPHABET_27
BIGBASE = len(ALPHABET)


DictProxyType = type(object.__dict__)


def make_hash(o):
    r"""
    Makes a hash from a dictionary, list, tuple or set to any level, that
    contains only other hashable types (including any lists, tuples, sets, and
    dictionaries). In the case where other kinds of objects (like classes) need
    to be hashed, pass in a collection of object attributes that are pertinent.
    For example, a class can be hashed in this fashion:

    make_hash([cls.__dict__, cls.__name__])

    A function can be hashed like so:

    make_hash([fn.__dict__, fn.__code__])

    References:
        http://stackoverflow.com/questions/5884066/hashing-a-python-dictionary
    """
    if type(o) == DictProxyType:
        o2 = {}
        for k, v in o.items():
            if not k.startswith("__"):
                o2[k] = v
            o = o2
    if isinstance(o, (set, tuple, list)):
        return tuple([make_hash(e) for e in o])
    elif not isinstance(o, dict):
        return hash(o)
    new_o = copy.deepcopy(o)
    for k, v in new_o.items():
        new_o[k] = make_hash(v)
    return hash(tuple(frozenset(sorted(new_o.items()))))


def hashstr_arr27(arr, lbl, alphabet=ALPHABET_27, **kwargs):
    return hashstr_arr(arr, lbl, alphabet=alphabet, **kwargs)


def hashstr27(data, alphabet=ALPHABET_27, **kwargs):
    return hashstr(data, alphabet=alphabet, **kwargs)


def hashstr_arr(arr, lbl='arr', pathsafe=False, **kwargs):
    r"""
    Args:
        arr (ndarray):
        lbl (str): (default = 'arr')
        pathsafe (bool): (default = False)

    Returns:
        str: arr_hashstr

    CommandLine:
        python -m utool.util_hash --test-hashstr_arr
        python -m utool.util_hash hashstr_arr:2

    Example0:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> import numpy as np
        >>> arr = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64)
        >>> lbl = 'arr'
        >>> kwargs = {}
        >>> pathsafe = False
        >>> arr_hashstr = hashstr_arr(arr, lbl, pathsafe, alphabet=ALPHABET_27)
        >>> result = ('arr_hashstr = %s' % (str(arr_hashstr),))
        >>> print(result)
        arr_hashstr = arr((2,3)daukyreqnhfejkfs)

    Example1:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> import numpy as np
        >>> arr = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64)
        >>> kwargs = {}
        >>> lbl = 'arr'
        >>> pathsafe = True
        >>> arr_hashstr = hashstr_arr(arr, lbl, pathsafe, alphabet=ALPHABET_27)
        >>> result = ('arr_hashstr = %s' % (str(arr_hashstr),))
        >>> print(result)
        arr_hashstr = arr-_2,3_daukyreqnhfejkfs-
    """
    if isinstance(arr, list):
        arr = tuple(arr)  # force arrays into a tuple for hashability
        # TODO: maybe for into numpy array instead? tuples might have problems
    if pathsafe:
        lbrace1, rbrace1, lbrace2, rbrace2 = '_', '_', '-', '-'
    else:
        lbrace1, rbrace1, lbrace2, rbrace2 = '(', ')', '(', ')'
    if isinstance(arr, tuple):
        arr_shape = lbrace1 + str(len(arr)) + rbrace1
    else:
        # Arr should be an ndarray here. append info about the ndarray
        arr_shape = lbrace1 + ','.join(list(map(str, arr.shape))) + rbrace1
    arr_hashstr_ = hashstr(arr, **kwargs)
    arr_hashstr = ''.join([lbl, lbrace2, arr_shape, arr_hashstr_, rbrace2])
    return arr_hashstr


def hashid_arr(arr, label='arr', hashlen=16):
    """ newer version of hashstr_arr2 """
    hashstr = hash_data(arr)[0:hashlen]
    if isinstance(arr, (list, tuple)):
        shapestr = len(arr)
    else:
        shapestr = ','.join(list(map(str, arr.shape)))
    hashid = '{}-{}-{}'.format(label, shapestr, hashstr)
    return hashid

if six.PY2:
    stringlike = (basestring, bytes)  # NOQA
if six.PY3:
    stringlike = (str, bytes)  # NOQA


def _covert_to_hashable(data):
    r"""
    Args:
        data (?):

    Returns:
        ?:

    CommandLine:
        python -m utool.util_hash _covert_to_hashable

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> from utool.util_hash import _covert_to_hashable  # NOQA
        >>> import utool as ut
        >>> data = np.array([1], dtype=np.int64)
        >>> result = _covert_to_hashable(data)
        >>> print(result)
    """
    if isinstance(data, six.binary_type):
        hashable = data
        prefix = b'TXT'
    elif util_type.HAVE_NUMPY and isinstance(data, np.ndarray):
        if data.dtype.kind == 'O':
            msg = '[ut] hashing ndarrays with dtype=object is unstable'
            warnings.warn(msg, RuntimeWarning)
            hashable = data.dumps()
        else:
            hashable = data.tobytes()
        prefix = b'NDARR'
    elif isinstance(data, six.text_type):
        # convert unicode into bytes
        hashable = data.encode('utf-8')
        prefix = b'TXT'
    elif isinstance(data, uuid.UUID):
        hashable = data.bytes
        prefix = b'UUID'
    elif isinstance(data, int):
        # warnings.warn('[util_hash] Hashing ints is slow, numpy is prefered')
        hashable = _int_to_bytes(data)
        # hashable = data.to_bytes(8, byteorder='big')
        prefix = b'INT'
    # elif isinstance(data, float):
    #     hashable = repr(data).encode('utf8')
    #     prefix = b'FLT'
    elif util_type.HAVE_NUMPY and isinstance(data, np.int64):
        return _covert_to_hashable(int(data))
    elif util_type.HAVE_NUMPY and isinstance(data, np.float64):
        a, b = float(data).as_integer_ratio()
        hashable = (a.to_bytes(8, byteorder='big') +
                    b.to_bytes(8, byteorder='big'))
        prefix = b'FLOAT'
    else:
        raise TypeError('unknown hashable type=%r' % (type(data)))
        # import bencode
        # hashable = bencode.Bencoder.encode(data).encode('utf-8')
        # prefix = b'BEN'
    prefix = b''
    return prefix, hashable


def _update_hasher(hasher, data):
    """
    This is the clear winner over the generate version.
    Used by hash_data

    Ignore:
        import utool
        rng = np.random.RandomState(0)
        # str1 = rng.rand(0).dumps()
        str1 = b'SEP'
        str2 = rng.rand(10000).dumps()
        for timer in utool.Timerit(100, label='twocall'):
            hasher = hashlib.sha256()
            with timer:
                hasher.update(str1)
                hasher.update(str2)
        a = hasher.hexdigest()
        for timer in utool.Timerit(100, label='concat'):
            hasher = hashlib.sha256()
            with timer:
                hasher.update(str1 + str2)
        b = hasher.hexdigest()
        assert a == b
        # CONCLUSION: Faster to concat in case of prefixes and seps

        nested_data = {'1': [rng.rand(100), '2', '3'],
                       '2': ['1', '2', '3', '4', '5'],
                       '3': [('1', '2'), ('3', '4'), ('5', '6')]}
        data = list(nested_data.values())


        for timer in utool.Timerit(1000, label='cat-generate'):
            hasher = hashlib.sha256()
            with timer:
                hasher.update(b''.join(_bytes_generator(data)))

        for timer in utool.Timerit(1000, label='inc-generate'):
            hasher = hashlib.sha256()
            with timer:
                for b in _bytes_generator(data):
                    hasher.update(b)

        for timer in utool.Timerit(1000, label='inc-generate'):
            hasher = hashlib.sha256()
            with timer:
                for b in _bytes_generator(data):
                    hasher.update(b)

        for timer in utool.Timerit(1000, label='chunk-inc-generate'):
            hasher = hashlib.sha256()
            import ubelt as ub
            with timer:
                for chunk in ub.chunks(_bytes_generator(data), 5):
                    hasher.update(b''.join(chunk))

        for timer in utool.Timerit(1000, label='inc-update'):
            hasher = hashlib.sha256()
            with timer:
                _update_hasher(hasher, data)

        data = ut.lorium_ipsum()
        hash_data(data)
        ut.hashstr27(data)
        %timeit hash_data(data)
        %timeit ut.hashstr27(repr(data))

        for timer in utool.Timerit(100, label='twocall'):
            hasher = hashlib.sha256()
            with timer:
                hash_data(data)

        hasher = hashlib.sha256()
        hasher.update(memoryview(np.array([1])))
        print(hasher.hexdigest())

        hasher = hashlib.sha256()
        hasher.update(np.array(['1'], dtype=object))
        print(hasher.hexdigest())

    """
    if isinstance(data, (tuple, list, zip)):
        needs_iteration = True
    elif (util_type.HAVE_NUMPY and isinstance(data, np.ndarray) and
          data.dtype.kind == 'O'):
        # ndarrays of objects cannot be hashed directly.
        needs_iteration = True
    else:
        needs_iteration = False

    if needs_iteration:
        # try to nest quickly without recursive calls
        SEP = b'SEP'
        iter_prefix = b'ITER'
        # if isinstance(data, tuple):
        #     iter_prefix = b'TUP'
        # else:
        #     iter_prefix = b'LIST'
        iter_ = iter(data)
        hasher.update(iter_prefix)
        try:
            for item in iter_:
                prefix, hashable = _covert_to_hashable(data)
                binary_data = SEP + prefix + hashable
                # b''.join([SEP, prefix, hashable])
                hasher.update(binary_data)
        except TypeError:
            # need to use recursive calls
            # Update based on current item
            _update_hasher(hasher, item)
            for item in iter_:
                # Ensure the items have a spacer between them
                hasher.update(SEP)
                _update_hasher(hasher, item)
    else:
        prefix, hashable = _covert_to_hashable(data)
        binary_data = prefix + hashable
        # b''.join([prefix, hashable])
        hasher.update(binary_data)


# def _bytes_generator(data):
#     # SLOWER METHOD
#     if isinstance(data, (tuple, list)):
#         # Ensure there is a iterable prefix with a spacer item
#         SEP = b'SEP'
#         iter_prefix = b'ITER'
#         # if isinstance(data, tuple):
#         #     iter_prefix = b'TUP'
#         # else:
#         #     iter_prefix = b'LIST'
#         iter_ = iter(data)
#         yield iter_prefix
#         try:
#             # try to nest quickly without recursive calls
#             for item in iter_:
#                 prefix, hashable = _covert_to_hashable(data)
#                 yield SEP
#                 yield prefix
#                 yield hashable
#         except TypeError:
#             # recover from failed item and then continue iterating using slow
#             # recursive calls
#             yield SEP
#             for bytes_ in _bytes_generator(item):
#                 yield bytes_
#             for item in iter_:
#                 yield SEP
#                 for bytes_ in _bytes_generator(item):
#                     yield bytes_
#     else:
#         prefix, hashable = _covert_to_hashable(data)
#         yield prefix
#         yield hashable


def b(x):
    return six.binary_type(six.b(x))


SEP_STR = '-'
SEP_BYTE = b(SEP_STR)


def freeze_hash_bytes(bytes_):
    import codecs
    hexstr = codecs.encode(bytes_, 'hex').decode('utf8')
    return hexstr


def combine_hashes(bytes_list, hasher=None):
    """
    Only works on bytes

    Example:
        >>> # DISABLE_DOCTEST
        >>> x = [b('1111'), b('2222')]
        >>> y = [b('11'), b('11'), b('22'), b('22')]
        >>> bytes_list = y
        >>> out1 = ut.combine_hashes(x, hashlib.sha1())
        >>> hasher = hashlib.sha1()
        >>> out2 = ut.combine_hashes(y, hasher)
        >>> bytes_ = out2
        >>> assert hasher.hexdigest() == freeze_hash_bytes(hasher.digest())
        >>> assert convert_bytes_to_bigbase(hasher.digest()) == convert_hexstr_to_bigbase(hasher.hexdigest())
        >>> assert out1 != out2
        >>> print('out1 = %r' % (out1,))
        >>> print('out2 = %r' % (out2,))
    """
    if hasher is None:
        hasher = hashlib.sha256()
    for b in bytes_list:
        hasher.update(b)
        hasher.update(SEP_BYTE)
    return hasher.digest()


@profile
def hash_data(data, hashlen=None, alphabet=None):
    r"""
    Get a unique hash depending on the state of the data.

    Args:
        data (object): any sort of loosely organized data
        hashlen (None): (default = None)
        alphabet (None): (default = None)

    Returns:
        str: text -  hash string

    CommandLine:
        python -m utool.util_hash hash_data

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> import utool as ut
        >>> counter = [0]
        >>> failed = []
        >>> def check_hash(input_, want=None):
        >>>     count = counter[0] = counter[0] + 1
        >>>     got = ut.hash_data(input_)
        >>>     print('({}) {}'.format(count, got))
        >>>     if want is not None and not got.startswith(want):
        >>>         failed.append((got, input_, count, want))
        >>> check_hash('1', 'wuvrng')
        >>> check_hash(['1'], 'dekbfpby')
        >>> check_hash(tuple(['1']), 'dekbfpby')
        >>> check_hash(b'12', 'marreflbv')
        >>> check_hash([b'1', b'2'], 'nwfs')
        >>> check_hash(['1', '2', '3'], 'arfrp')
        >>> check_hash(['1', np.array([1,2,3]), '3'], 'uyqwcq')
        >>> check_hash('123', 'ehkgxk')
        >>> check_hash(zip([1, 2, 3], [4, 5, 6]), 'mjcpwa')
        >>> import numpy as np
        >>> rng = np.random.RandomState(0)
        >>> check_hash(rng.rand(100000), 'bdwosuey')
        >>> for got, input_, count, want in failed:
        >>>     print('failed {} on {}'.format(count, input_))
        >>>     print('got={}, want={}'.format(got, want))
        >>> assert not failed
    """
    if alphabet is None:
        alphabet = ALPHABET_27
    if hashlen is None:
        hashlen = HASH_LEN2
    if isinstance(data, stringlike) and len(data) == 0:
        # Make a special hash for empty data
        text = (alphabet[0] * hashlen)
    else:
        hasher = hashlib.sha512()
        _update_hasher(hasher, data)
        # Get a 128 character hex string
        text = hasher.hexdigest()
        # Shorten length of string (by increasing base)
        hashstr2 = convert_hexstr_to_bigbase(text, alphabet, bigbase=len(alphabet))
        # Truncate
        text = hashstr2[:hashlen]
        return text


def digest_data(data, alg='sha256'):
    hasher = {
        'md5'    : hashlib.md5,
        'sha1'   : hashlib.sha1,
        'sha256' : hashlib.sha256,
    }[alg]()
    _update_hasher(hasher, data)
    return hasher.digest()


def hashstr(data, hashlen=HASH_LEN, alphabet=ALPHABET):
    """
    python -c "import utool as ut; print(ut.hashstr('abcd'))"

    Args:
        data (hashable):
        hashlen (int): (default = 16)
        alphabet (list): list of characters:

    Returns:
        str: hashstr

    CommandLine:
        python -m utool.util_hash --test-hashstr
        python3 -m utool.util_hash --test-hashstr
        python3 -m utool.util_hash --test-hashstr:2
        python -m utool.util_hash hashstr:3
        python3 -m utool.util_hash hashstr:3


    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> data = 'foobar'
        >>> hashlen = 16
        >>> alphabet = ALPHABET_41
        >>> text = hashstr(data, hashlen, alphabet)
        >>> result = ('text = %s' % (str(text),))
        >>> print(result)
        text = mi5yum60mbxhyp+x

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> data = ''
        >>> hashlen = 16
        >>> alphabet = ALPHABET_41
        >>> text = hashstr(data, hashlen, alphabet)
        >>> result = ('text = %s' % (str(text),))
        >>> print(result)
        text = 0000000000000000

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> import numpy as np
        >>> data = np.array([1, 2, 3])
        >>> hashlen = 16
        >>> alphabet = ALPHABET_41
        >>> text = hashstr(data, hashlen, alphabet)
        >>> result = ('text = %s' % (str(text),))
        >>> print(result)
        text = z5lqw0bzt4dmb9yy

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> import numpy as np
        >>> from uuid import UUID
        >>> data = (UUID('7cd0197b-1394-9d16-b1eb-0d8d7a60aedc'), UUID('c76b54a5-adb6-7f16-f0fb-190ab99409f8'))
        >>> hashlen = 16
        >>> alphabet = ALPHABET_41
        >>> text = hashstr_arr(data, 'label')
        >>> result = ('text = %s' % (str(text),))
        >>> print(result)


    Ignore:
        >>> # DISABLE_DOCTEST
        >>> # UNSTABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> import numpy as np
        >>> data = np.array(['a', 'b'], dtype=object)
        >>> text = hashstr(data, alphabet=ALPHABET_27)
        >>> result = ('text = %s' % (str(text),))
        >>> print(result)

    Ignore:
        data = np.array(['a', 'b'], dtype=object)
        data.tobytes()
        data = np.array(['a', 'b'])
        data = ['a', 'b']
        data = np.array([1, 2, 3])
        import hashlib
        from six.moves import cPickle as pickle
        pickle.dumps(data, protocol=2)

        python2 -c "import hashlib, numpy; print(hashlib.sha1('ab').hexdigest())"
        python3 -c "import hashlib, numpy; print(hashlib.sha1('ab'.encode('utf8')).hexdigest())"

        python2 -c "import hashlib, numpy; print(hashlib.sha1('ab').hexdigest())"
        python3 -c "import hashlib, numpy; print(hashlib.sha1(b'ab').hexdigest())"

        python2 -c "import hashlib, numpy; print(hashlib.sha1(numpy.array([1, 2])).hexdigest())"
        python3 -c "import hashlib, numpy; print(hashlib.sha1(numpy.array([1, 2])).hexdigest())"

        # TODO: numpy arrays of strings must be encoded to bytes first in python3
        python2 -c "import hashlib, numpy; print(hashlib.sha1(numpy.array(['a', 'b'])).hexdigest())"
        python3 -c "import hashlib, numpy; print(hashlib.sha1(numpy.array([b'a', b'b'])).hexdigest())"

        python -c "import hashlib, numpy; print(hashlib.sha1(numpy.array(['a', 'b'], dtype=object)).hexdigest())"
        python -c "import hashlib, numpy; print(hashlib.sha1(numpy.array(['a', 'b'], dtype=object)).hexdigest())"
    """
    if util_type.HAVE_NUMPY and isinstance(data, np.ndarray):
        if data.dtype.kind == 'O':
            msg = '[ut] hashing ndarrays with dtype=object is unstable'
            warnings.warn(msg, RuntimeWarning)
            # but tobytes is ok, but differs between python 2 and 3 for objects
            data = data.dumps()
            # data = data.tobytes()
    if isinstance(data, tuple):
        # should instead do
        if False:
            hasher = hashlib.sha512()
            items = data
            for item in items:
                if isinstance(item, uuid.UUID):
                    hasher.update(item.bytes)
                else:
                    hasher.update(item)
            text = hasher.hexdigest()
            hashstr2 = convert_hexstr_to_bigbase(text, alphabet, bigbase=len(alphabet))
            # Truncate
            text = hashstr2[:hashlen]
            return text
        else:
            msg = '[ut] hashing tuples with repr is not a good idea. FIXME'
            # warnings.warn(msg, RuntimeWarning)
            data = repr(data)  # Hack?

    # convert unicode into raw bytes
    if isinstance(data, six.text_type):
        data = data.encode('utf-8')

    if isinstance(data, stringlike) and len(data) == 0:
        # Make a special hash for empty data
        text = (alphabet[0] * hashlen)
    else:
        # Get a 128 character hex string
        text = hashlib.sha512(data).hexdigest()
        # Shorten length of string (by increasing base)
        hashstr2 = convert_hexstr_to_bigbase(text, alphabet, bigbase=len(alphabet))
        # Truncate
        text = hashstr2[:hashlen]
    return text

r"""
def valid_filename_ascii_chars():
    # Find invalid chars
    ntfs_inval = '< > : " / \ | ? *'.split(' ')
    other_inval = [' ', '\'', '.']
    #case_inval = map(chr, range(97, 123))
    case_inval = map(chr, range(65, 91))
    invalid_chars = set(ntfs_inval + other_inval + case_inval)
    # Find valid chars
    valid_chars = []
    for index in range(32, 127):
        char = chr(index)
        if not char in invalid_chars:
            print index, chr(index)
            valid_chars.append(chr(index))
    return valid_chars
valid_filename_ascii_chars()
"""

if six.PY3:
    def _int_to_bytes(int_):
        length = max(4, int_.bit_length())
        bytes_ = int_.to_bytes(length, byteorder='big')
        # bytes_ = int_.to_bytes(4, byteorder='big')
        # int_.to_bytes(8, byteorder='big')  # TODO: uncomment
        return bytes_

    def _bytes_to_int(bytes_):
        int_ = int.from_bytes(bytes_, 'big')
        return int_
else:
    def _py2_to_bytes(int_, length, byteorder='big'):
        h = '%x' % int_
        s = ('0' * (len(h) % 2) + h).zfill(length * 2).decode('hex')
        bytes_ =  s if byteorder == 'big' else s[::-1]
        return bytes_

    import codecs
    def _int_to_bytes(int_):
        length = max(4, int_.bit_length())
        bytes_ = _py2_to_bytes(int_, length, 'big')
        # bytes_ = struct.pack('>i', int_)
        return bytes_

    def _bytes_to_int(bytes_):
        int_ = int(codecs.encode(bytes_, 'hex'), 16)
        # int_ = struct.unpack('>i', bytes_)[0]
        # int_ = struct.unpack_from('>L', bytes_)[0]
        return int_


def _test_int_byte_conversion():
    import itertools as it
    import utool as ut
    inputs = list(it.chain(
        range(0, 10),
        (2 ** i for i in range(0, 256, 32)),
        (2 ** i + 1 for i in range(0, 256, 32)),
    ))
    for int_0 in inputs:
        print('---')
        print('int_0 = %s' % (ut.repr2(int_0),))
        bytes_ = _int_to_bytes(int_0)
        int_ = _bytes_to_int(bytes_)
        print('bytes_ = %s' % (ut.repr2(bytes_),))
        print('int_ = %s' % (ut.repr2(int_),))
        assert int_ == int_0


def convert_bytes_to_bigbase(bytes_, alphabet=ALPHABET_27):
    r"""
    Args:
        bytes_ (bytes):

    Returns:
        str:

    Ignore:

    CommandLine:
        python -m utool.util_hash convert_bytes_to_bigbase

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> import utool as ut
        >>> bytes_ = b('9999999999999999999999999999999999')
        >>> alphabet = ALPHABET_27
        >>> result = convert_bytes_to_bigbase(bytes_, alphabet)
        >>> print(result)
        fervudwhpustklnptklklcgswbmvtustqocdpgiwkgrvwytvneardkpytd
    """
    x = _bytes_to_int(bytes_)
    if x == 0:
        return '0'
    sign = 1 if x > 0 else -1
    x *= sign
    digits = []
    bigbase = len(alphabet)
    while x:
        digits.append(alphabet[x % bigbase])
        x //= bigbase
    if sign < 0:
        digits.append('-')
        digits.reverse()
    newbase_str = ''.join(digits)
    return newbase_str


def convert_hexstr_to_bigbase(hexstr, alphabet=ALPHABET, bigbase=BIGBASE):
    r"""
    Packs a long hexstr into a shorter length string with a larger base

    Ignore:
        # Determine the length savings with lossless conversion
        import sympy as sy
        consts = dict(hexbase=16, hexlen=256, bigbase=27)
        symbols = sy.symbols('hexbase, hexlen, bigbase, newlen')
        haexbase, hexlen, bigbase, newlen = symbols
        eqn = sy.Eq(16 ** hexlen,  bigbase ** newlen)
        newlen_ans = sy.solve(eqn, newlen)[0].subs(consts).evalf()
        print('newlen_ans = %r' % (newlen_ans,))

        # for a 27 char alphabet we can get 216
        print('Required length for lossless conversion len2 = %r' % (len2,))

        def info(base, len):
            bits = base ** len
            print('base = %r' % (base,))
            print('len = %r' % (len,))
            print('bits = %r' % (bits,))
        info(16, 256)
        info(27, 16)
        info(27, 64)
        info(27, 216)
    """
    x = int(hexstr, 16)  # first convert to base 16
    if x == 0:
        return '0'
    sign = 1 if x > 0 else -1
    x *= sign
    digits = []
    while x:
        digits.append(alphabet[x % bigbase])
        x //= bigbase
    if sign < 0:
        digits.append('-')
        digits.reverse()
    newbase_str = ''.join(digits)
    return newbase_str


def hashstr_md5(data):
    """
    Ignore:
        >>> # xdoctest: +SKIP
        >>> from utool.util_hash import *  # NOQA
        >>> fpath = ut.grab_test_imgpath('patsy.jpg')
        %timeit ut.get_file_hash(fpath, hasher=hashlib.sha1())
        %timeit ut.get_file_hash(fpath, hasher=hashlib.md5())

    """
    text = hashlib.md5(data).hexdigest()
    #bin(int(my_hexdata, scale))
    return text


def hashstr_sha1(data, base10=False):
    text = hashlib.sha1(data).hexdigest()
    if base10:
        text = int("0x" + text, 0)
    return text


def get_file_hash(fpath, blocksize=65536, hasher=None, stride=1,
                  hexdigest=False):
    r"""
    For better hashes use hasher=hashlib.sha256, and keep stride=1

    Args:
        fpath (str):  file path string
        blocksize (int): 2 ** 16. Affects speed of reading file
        hasher (None):  defaults to sha1 for fast (but insecure) hashing
        stride (int): strides > 1 skip data to hash, useful for faster
                      hashing, but less accurate, also makes hash dependant on
                      blocksize.

    References:
        http://stackoverflow.com/questions/3431825/generating-a-md5-checksum-of-a-file
        http://stackoverflow.com/questions/5001893/when-should-i-use-sha-1-and-when-should-i-use-sha-2

    CommandLine:
        python -m utool.util_hash --test-get_file_hash
        python -m utool.util_hash --test-get_file_hash:0
        python -m utool.util_hash --test-get_file_hash:1

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> fpath = ut.grab_test_imgpath('patsy.jpg')
        >>> #blocksize = 65536  # 2 ** 16
        >>> blocksize = 2 ** 16
        >>> hasher = None
        >>> stride = 1
        >>> hashbytes_20 = get_file_hash(fpath, blocksize, hasher, stride)
        >>> result = repr(hashbytes_20)
        >>> print(result)
        '7\x07B\x0eX<sRu\xa2\x90P\xda\xb2\x84?\x81?\xa9\xd9'

        '\x13\x9b\xf6\x0f\xa3QQ \xd7"$\xe9m\x05\x9e\x81\xf6\xf2v\xe4'

        '\x16\x00\x80Xx\x8c-H\xcdP\xf6\x02\x9frl\xbf\x99VQ\xb5'

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> #fpath = ut.grab_file_url('http://en.wikipedia.org/wiki/List_of_comets_by_type')
        >>> fpath = ut.unixjoin(ut.ensure_app_resource_dir('utool'), 'tmp.txt')
        >>> ut.write_to(fpath, ut.lorium_ipsum())
        >>> blocksize = 2 ** 3
        >>> hasher = None
        >>> stride = 2
        >>> hashbytes_20 = get_file_hash(fpath, blocksize, hasher, stride)
        >>> result = repr(hashbytes_20)
        >>> print(result)
        '5KP\xcf>R\xf6\xffO:L\xac\x9c\xd3V+\x0e\xf6\xe1n'

    Ignore:
        file_ = open(fpath, 'rb')
    """
    if hasher is None:
        hasher = hashlib.sha1()
    with open(fpath, 'rb') as file_:
        buf = file_.read(blocksize)
        while len(buf) > 0:
            hasher.update(buf)
            if stride > 1:
                file_.seek(blocksize * (stride - 1), 1)  # skip blocks
            buf = file_.read(blocksize)
        if hexdigest:
            return hasher.hexdigest()
        else:
            return hasher.digest()


def write_hash_file(fpath, hash_tag='md5', recompute=False):
    r""" Creates a hash file for each file in a path

    CommandLine:
        python -m utool.util_hash --test-write_hash_file

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> import utool as ut
        >>> from utool.util_hash import *  # NOQA
        >>> fpath = ut.grab_test_imgpath('patsy.jpg')
        >>> write_hash_file(fpath, 'md5')
    """
    hash_dict = {
        'md5'    : hashlib.md5(),
        'sha1'   : hashlib.sha1(),
        'sha256' : hashlib.sha256(),
    }
    message = "Unrecognized hashing function.  Use 'md5', 'sha1', or 'sha256"
    assert hash_tag in hash_dict, message
    if fpath.endswith('.%s' % (hash_tag, )):
        # No need to compute hashes on hashes
        return
    # Get hash path
    hash_fpath = '%s.%s' % (fpath, hash_tag, )
    if os.path.exists(hash_fpath) and not recompute:
        return
    # Assert this is a file
    file_type = util_path.get_path_type(fpath)
    if file_type == 'file':
        # Compute hash
        hasher = hash_dict[hash_tag]
        hash_local = get_file_hash(fpath, hasher=hasher, hexdigest=True)
        print('[utool] Adding:', fpath, hash_local)
        with open(hash_fpath, 'w') as hash_file:
            hash_file.write(hash_local)
        return hash_fpath


def write_hash_file_for_path(path, recompute=False):
    r""" Creates a hash file for each file in a path

    CommandLine:
        python -m utool.util_hash --test-write_hash_file_for_path

    Ignore:
        >>> # DISABLE_DOCTEST
        >>> import os
        >>> import utool as ut
        >>> from utool.util_hash import *  # NOQA
        >>> fpath = ut.grab_test_imgpath('patsy.jpg')
        >>> path, _ = os.path.split(fpath)
        >>> hash_fpath_list = write_hash_file_for_path(path)
        >>> for hash_fpath in hash_fpath_list:
        >>>     assert os.path.exists(hash_fpath)
        >>>     ut.delete(hash_fpath)
    """
    hash_fpath_list = []
    for root, dname_list, fname_list in os.walk(path):
        for fname in sorted(fname_list):
            # fpath = os.path.join(path, fname)
            fpath = os.path.join(root, fname)
            hash_fpath = write_hash_file(fpath, recompute=recompute)
            if hash_fpath is not None:
                hash_fpath_list.append(hash_fpath)
    return hash_fpath_list


def get_file_uuid(fpath, hasher=None, stride=1):
    """ Creates a uuid from the hash of a file
    """
    if hasher is None:
        hasher = hashlib.sha1()  # 20 bytes of output
        #hasher = hashlib.sha256()  # 32 bytes of output
    # sha1 produces a 20 byte hash
    hashbytes_20 = get_file_hash(fpath, hasher=hasher, stride=stride)
    # sha1 produces 20 bytes, but UUID requires 16 bytes
    hashbytes_16 = hashbytes_20[0:16]
    uuid_ = uuid.UUID(bytes=hashbytes_16)
    return uuid_


def image_uuid(pil_img):
    """
    UNSAFE: DEPRICATE: JPEG IS NOT GAURENTEED TO PRODUCE CONSITENT VALUES ON

    MULTIPLE MACHINES image global unique id

    References:
        http://stackoverflow.com/questions/23565889/jpeg-images-have-different-pixel-values-across-multiple-devices
    """
    print('WARNING DO NOT USE utool.util_hash.image_uuid UNSAFE AND DEPRICATED')
    # Get the bytes of the image
    img_bytes_ = pil_img.tobytes()
    uuid_ = hashable_to_uuid(img_bytes_)
    return uuid_


def augment_uuid(uuid_, *hashables):
    #from six.moves import reprlib
    #uuidhex_data   = uuid_.get_bytes()
    uuidhex_data   = uuid_.bytes
    #hashable_str    = ''.join(map(repr, hashables))
    # Python 2 and 3 diverge here because repr returns
    # ascii data in python2 and unicode text in python3
    # it would be nice to
    # warnings.warn('[ut] should not use repr when hashing', RuntimeWarning)
    def tmprepr(x):
        y = repr(x)
        # hack to remove u prefix
        if isinstance(x, six.string_types):
            if y.startswith('u'):
                y = y[1:]
        return y
    if six.PY2:
        hashable_text = ''.join(map(tmprepr, hashables))
        hashable_data = hashable_text.encode('utf-8')
        #hashable_data = b''.join(map(bytes, hashables))
    elif six.PY3:
        hashable_text    = ''.join(map(tmprepr, hashables))
        hashable_data = hashable_text.encode('utf-8')
        #hashable_data = b''.join(map(bytes, hashables))
    augmented_data   = uuidhex_data + hashable_data
    augmented_uuid_ = hashable_to_uuid(augmented_data)
    return augmented_uuid_


@profile
def combine_uuids(uuids, ordered=True, salt=''):
    """
    Creates a uuid that specifies a group of UUIDS

    Args:
        uuids (list): list of uuid objects
        ordered (bool): if False uuid order changes the resulting combined uuid
            otherwise the uuids are considered an orderless set
        salt (str): salts the resulting hash

    Returns:
        uuid.UUID: combined uuid

    CommandLine:
        python -m utool.util_hash --test-combine_uuids

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> import utool as ut
        >>> uuids = [hashable_to_uuid('one'), hashable_to_uuid('two'),
        >>>          hashable_to_uuid('three')]
        >>> combo1 = combine_uuids(uuids, ordered=True)
        >>> combo2 = combine_uuids(uuids[::-1], ordered=True)
        >>> combo3 = combine_uuids(uuids, ordered=False)
        >>> combo4 = combine_uuids(uuids[::-1], ordered=False)
        >>> result = ut.repr4([combo1, combo2, combo3, combo4], nobr=True)
        >>> print(result)
        UUID('83ee781f-8646-ccba-0ed8-13842825c12a'),
        UUID('52bbb33f-612e-2ab8-a62c-2f46e5b1edc8'),
        UUID('945cadab-e834-e581-0f74-62f106d20d81'),
        UUID('945cadab-e834-e581-0f74-62f106d20d81'),

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> import utool as ut
        >>> uuids = [uuid.UUID('5ff6b34e-7d8f-ef32-5fad-489266acd2ae'),
        >>>          uuid.UUID('f2400146-ec12-950b-1489-668228e155a8'),
        >>>          uuid.UUID('037d6f31-8c73-f961-1fe4-d616442a1e86'),
        >>>          uuid.UUID('ca45d6e2-e648-09cc-a49e-e71c6fa3b3f3')]
        >>> ordered = True
        >>> salt = u''
        >>> result = combine_uuids(uuids, ordered, salt)
        >>> print(result)
        1dabc66b-b564-676a-99b4-5cae7a9e7294
    """
    if len(uuids) == 0:
        return get_zero_uuid()
    elif len(uuids) == 1:
        return uuids[0]
    else:
        if not ordered:
            uuids = sorted(uuids)
        sep_str = '-'
        sep_byte = six.binary_type(six.b(sep_str))
        pref = six.binary_type(six.b('{}{}{}'.format(salt, sep_str, len(uuids))))
        combined_bytes = pref + sep_byte.join([u.bytes for u in uuids])
        combined_uuid = hashable_to_uuid(combined_bytes)
        return combined_uuid


if six.PY3:
    def _ensure_hashable_bytes(hashable_):
        # If hashable_ is text (python3)
        if isinstance(hashable_, bytes):
            return hashable_
        elif isinstance(hashable_, str):
            return hashable_.encode('utf-8')
        elif isinstance(hashable_, int):
            return hashable_.to_bytes(4, byteorder='big')
            # return int_.to_bytes(8, byteorder='big')  # TODO: uncomment
        elif isinstance(hashable_, (list, tuple)):
            return str(hashable_).encode('utf-8')
        else:
            return hashable_
elif six.PY2:
    import struct
    def _ensure_hashable_bytes(hashable_):
        # If hashable_ is data (python2)
        if isinstance(hashable_, bytes):
            return hashable_
        elif isinstance(hashable_, str):
            return hashable_.encode('utf-8')
        elif isinstance(hashable_, int):
            return struct.pack('>i', hashable_)
        elif isinstance(hashable_, (list, tuple)):
            return str(hashable_).encode('utf-8')
        else:
            return bytes(hashable_)


def hashable_to_uuid(hashable_):
    """
    TODO: ensure that python2 and python3 agree on hashes of the same
    information

    Args:
        hashable_ (hashable): hashables are bytes-like objects
           An object that supports the Buffer Protocol, like bytes, bytearray
           or memoryview. Bytes-like objects can be used for various operations
           that expect binary data, such as compression, saving to a binary
           file or sending over a socket. Some operations need the binary data
           to be mutable, in which case not all bytes-like objects can apply.

    Returns:
        UUID: uuid_

    CommandLine:
        python -m utool.util_hash --test-hashable_to_uuid

    Ignore:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_hash import *  # NOQA
        >>> import utool as ut
        >>> hashables = [
        >>>     'foobar',
        >>>     'foobar'.encode('utf-8'),
        >>>     u'foobar',
        >>>     10,
        >>>     [1, 2, 3],
        >>> ]
        >>> uuids = []
        >>> for hashable_ in hashables:
        >>>     uuid_ = hashable_to_uuid(hashable_)
        >>>     uuids.append(uuid_)
        >>> result = ut.repr4(ut.lmap(str, uuids), strvals=True, nobr=True)
        >>> print(result)
        8843d7f9-2416-211d-e9eb-b963ff4ce281,
        8843d7f9-2416-211d-e9eb-b963ff4ce281,
        8843d7f9-2416-211d-e9eb-b963ff4ce281,
        e864ece8-8880-43b6-7277-c8b2cefe96ad,
        a01eda32-e4e0-b139-3274-e91d1b3e9ecf,
    """
    bytes_ = _ensure_hashable_bytes(hashable_)
    try:
        bytes_sha1 = hashlib.sha1(bytes_)
    except TypeError:
        print('hashable_ = %r' % (hashable_,))
        print('bytes_ = %r' % (bytes_,))
        raise
    # Digest them into a hash
    hashbytes_20 = bytes_sha1.digest()
    hashbytes_16 = hashbytes_20[0:16]
    uuid_ = uuid.UUID(bytes=hashbytes_16)
    return uuid_


def random_uuid():
    return uuid.uuid4()


def random_nonce(length=64, alphabet=None):
    """
    returns a random string of len=<length> from <alphabet>
    I have no idea why this is named random_nonce
    """
    assert length > 0
    if alphabet is None:
        alphabet = ALPHABET_16
    return ''.join( [alphabet[random.randint(0, len(alphabet) - 1)] for _ in range(length)] )


def get_zero_uuid():
    return uuid.UUID('00000000-0000-0000-0000-000000000000')

# Cleanup namespace
# del ALPHABET_41
del ALPHABET_54


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_hash
        python -m utool.util_hash --allexamples
        python3 -m utool.util_hash --allexamples
        python -m utool.util_hash --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
DEPRICATE
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import six
import re

MAX_VALSTR = -1


class AbstractPrintable(object):
    """
    A base class that prints its attributes instead of the memory address
    """

    def __init__(self, child_print_exclude=[]):
        self._printable_exclude = ['_printable_exclude'] + child_print_exclude

    def __str__(self):
        head = printableType(self)
        body = self.get_printable(type_bit=True)
        body = re.sub('\n *\n *\n', '\n\n', body)
        return head + ('\n' + body).replace('\n', '\n    ')

    def printme(self):
        print(str(self))

    def printme3(self):
        print(self.get_printable())

    def printme2(self,
                 type_bit=True,
                 print_exclude_aug=[],
                 val_bit=True,
                 max_valstr=MAX_VALSTR,
                 justlength=True):
        to_print = self.get_printable(type_bit=type_bit,
                                      print_exclude_aug=print_exclude_aug,
                                      val_bit=val_bit,
                                      max_valstr=max_valstr,
                                      justlength=justlength)
        print(to_print)

    def get_printable(self,
                      type_bit=True,
                      print_exclude_aug=[],
                      val_bit=True,
                      max_valstr=MAX_VALSTR,
                      justlength=False):
        from utool.util_str import truncate_str
        body = ''
        attri_list = []
        exclude_key_list = list(self._printable_exclude) + list(print_exclude_aug)
        for (key, val) in six.iteritems(self.__dict__):
            try:
                if key in exclude_key_list:
                    continue
                namestr = str(key)
                typestr = printableType(val, name=key, parent=self)
                if not val_bit:
                    attri_list.append((typestr, namestr, '<ommited>'))
                    continue
                valstr = printableVal(val, type_bit=type_bit, justlength=justlength)
                valstr = truncate_str(valstr, maxlen=max_valstr, truncmsg=' \n ~~~ \n ')
                #if len(valstr) > max_valstr and max_valstr > 0:
                #    pos1 =  max_valstr // 2
                #    pos2 = -max_valstr // 2
                #    valstr = valstr[0:pos1] + ' \n ~~~ \n ' + valstr[pos2: - 1]
                attri_list.append((typestr, namestr, valstr))
            except Exception as ex:
                print('[printable] ERROR %r' % ex)
                print('[printable] ERROR key = %r' % key)
                print('[printable] ERROR val = %r' % val)
                try:
                    print('[printable] ERROR valstr = %r' % valstr)
                except Exception:
                    pass
                raise
        attri_list.sort()
        for (typestr, namestr, valstr) in attri_list:
            entrytail = '\n' if valstr.count('\n') <= 1 else '\n\n'
            typestr2 = typestr + ' ' if type_bit else ''
            body += typestr2 + namestr + ' = ' + valstr + entrytail
        return body

    def format_printable(self, type_bit=False, indstr='  * '):
        _printable_str = self.get_printable(type_bit=type_bit)
        _printable_str = _printable_str.replace('\r', '\n')
        _printable_str = indstr + _printable_str.strip('\n').replace('\n', '\n' + indstr)
        return _printable_str


# - --------------

def printableType(val, name=None, parent=None):
    """
    Tries to make a nice type string for a value.
    Can also pass in a Printable parent object
    """
    import numpy as np
    if parent is not None and hasattr(parent, 'customPrintableType'):
        # Hack for non - trivial preference types
        _typestr = parent.customPrintableType(name)
        if _typestr is not None:
            return _typestr
    if isinstance(val, np.ndarray):
        info = npArrInfo(val)
        _typestr = info.dtypestr
    elif isinstance(val, object):
        _typestr = val.__class__.__name__
    else:
        _typestr = str(type(val))
        _typestr = _typestr.replace('type', '')
        _typestr = re.sub('[\'><]', '', _typestr)
        _typestr = re.sub('  *', ' ', _typestr)
        _typestr = _typestr.strip()
    return _typestr


def printableVal(val, type_bit=True, justlength=False):
    """
    Very old way of doing pretty printing. Need to update and refactor.
    DEPRICATE
    """
    from utool import util_dev
    # Move to util_dev
    # NUMPY ARRAY
    import numpy as np
    if type(val) is np.ndarray:
        info = npArrInfo(val)
        if info.dtypestr.startswith('bool'):
            _valstr = '{ shape:' + info.shapestr + ' bittotal: ' + info.bittotal + '}'
            # + '\n  |_____'
        elif info.dtypestr.startswith('float'):
            _valstr = util_dev.get_stats_str(val)
        else:
            _valstr = '{ shape:' + info.shapestr + ' mM:' + info.minmaxstr + ' }'  # + '\n  |_____'
    # String
    elif isinstance(val, six.text_type):  # NOQA
        _valstr = '\'%s\'' % val
    # List
    elif isinstance(val, list):
        if justlength or len(val) > 30:
            _valstr = 'len=' + str(len(val))
        else:
            _valstr = '[ ' + (', \n  '.join([str(v) for v in val])) + ' ]'
    # ??? isinstance(val, AbstractPrintable):
    elif hasattr(val, 'get_printable') and type(val) != type:
        _valstr = val.get_printable(type_bit=type_bit)
    elif isinstance(val, dict):
        _valstr = '{\n'
        for val_key in val.keys():
            val_val = val[val_key]
            _valstr += '  ' + str(val_key) + ' : ' + str(val_val) + '\n'
        _valstr += '}'
    else:
        _valstr = str(val)
    if _valstr.find('\n') > 0:  # Indent if necessary
        _valstr = _valstr.replace('\n', '\n    ')
        _valstr = '\n    ' + _valstr
    _valstr = re.sub('\n *$', '', _valstr)  # Replace empty lines
    return _valstr


def npArrInfo(arr):
    """
    OLD update and refactor
    """
    from utool.DynamicStruct import DynStruct
    info = DynStruct()
    info.shapestr  = '[' + ' x '.join([str(x) for x in arr.shape]) + ']'
    info.dtypestr  = str(arr.dtype)
    if info.dtypestr == 'bool':
        info.bittotal = 'T=%d, F=%d' % (sum(arr), sum(1 - arr))
    elif info.dtypestr == 'object':
        info.minmaxstr = 'NA'
    elif info.dtypestr[0] == '|':
        info.minmaxstr = 'NA'
    else:
        if arr.size > 0:
            info.minmaxstr = '(%r, %r)' % (arr.min(), arr.max())
        else:
            info.minmaxstr = '(None)'
    return info



# -*- coding: utf-8 -*-
# Utools for setup.py files
from __future__ import absolute_import, division, print_function
import sys
import textwrap
from os.path import exists, join, dirname, split, splitext
import os
from utool import util_cplat
from utool import util_path
from utool import util_io
from utool import util_str
#from utool import util_dev
from utool.util_dbg import printex
from utool._internal import meta_util_arg

VERBOSE = meta_util_arg.VERBOSE


class SetupManager(object):
    """
    Helps with writing setup.py
    """
    def __init__(self):
        self.cmdclass = {}

    def _register_command(self, name, func):
        import setuptools
        class _WrapCommand(setuptools.Command):
            """
            https://dankeder.com/posts/adding-custom-commands-to-setup-py/
            """
            description = name
            user_options = []
            def initialize_options(self):
                pass
            def finalize_options(self):
                pass
            def run(cmd):
                func()
        self.cmdclass[name] = _WrapCommand
        #self.cmdclass[name] = func
        #TmpCommand

    def register_command(self, name):
        import utool as ut
        if ut.is_funclike(name):
            func = name
            name = ut.get_funcname(func)
            self._register_command(name, func)
            return func
        else:
            def _wrap(func):
                self._register_command(name, func)
            return _wrap

    def get_cmdclass(self):
        cmdclass = self.cmdclass.copy()
        cmdclass.update(**get_cmdclass())
        return cmdclass


class SETUP_PATTERNS():
    clutter_pybuild = [
        '*.pyc',
        '*.pyo',
    ]
    clutter_cyth = [
        '_*_cyth.o',
        '_*_cyth_bench.py',
        'run_cyth_benchmarks.sh',
        '_*_cyth.c',
        '_*_cyth.pyd',
        '_*_cyth.pxd',
        '_*_cyth.html',
        '_*_cyth.pyx',
        '_*_cyth.so',
        '_*_cyth.dylib'
    ]
    chmod_test   = ['test_*.py']


def read_license(license_file):
    try:
        with open(license_file, 'r') as file_:
            firstline = file_.readline()
            license_type = firstline.replace('License', '').strip()
            return license_type
    except IOError:
        print('Warning no LICENCE file')
        return '???'


def build_pyo(project_dirs):
    pyexe = util_cplat.python_executable()
    for dir_ in project_dirs:
        #command_args = [pyexe, '-O', '-m', 'compileall', dir_ + '\*.py']
        command_args = [pyexe, '-O', '-m', 'compileall', dir_]
        #command = ' '.join(command_args)
        #os.system(command)
        util_cplat.shell(command_args)


def setup_chmod(setup_fpath, setup_dir, chmod_patterns):
    """ Gives files matching pattern the same chmod flags as setup.py """
    #st_mode = os.stat(setup_fpath).st_mode
    st_mode = 33277
    for pattern in chmod_patterns:
        for fpath in util_path.glob(setup_dir, pattern, recursive=True):
            print('[setup] chmod fpath=%r' % fpath)
            os.chmod(fpath, st_mode)


def assert_in_setup_repo(setup_fpath, name=''):
    """ pass in __file__ from setup.py """
    setup_dir, setup_fname = split(setup_fpath)
    cwd = os.getcwd()
    #repo_dname = split(setup_dir)[1]
    #print('cwd       = %r' % (cwd))
    #print('repo_dname = %r' % repo_dname)
    #print('setup_dir = %r' % (setup_dir))
    #print('setup_fname = %r' % (setup_fname))
    try:
        assert setup_fname == 'setup.py', 'name is not setup.py'
        #assert name == '' or repo_dname == name,
        ('name=%r' % name)
        assert cwd == setup_dir, 'cwd is not setup_dir'
        assert exists(setup_dir), 'setup dir does not exist'
        assert exists(join(setup_dir, 'setup.py')), 'setup.py does not exist'
    except AssertionError as ex:
        printex(ex, 'ERROR!: setup.py must be run from repository root')
        raise


def clean(setup_dir, clutter_patterns, clutter_dirs):
    print('[setup] clean()')

    clutter_patterns_ = [pat for pat in clutter_patterns if not pat.endswith('/')]
    _clutter_dirs = [pat[:-1] for pat in clutter_patterns if pat.endswith('/')]
    clutter_dirs_ = _clutter_dirs + clutter_dirs

    util_path.remove_files_in_dir(setup_dir,
                                  clutter_patterns_,
                                  recursive=True,
                                  verbose=VERBOSE)

    for dir_ in clutter_dirs_:
        util_path.delete(dir_, verbose=VERBOSE, print_exists=False)
        #util_path.remove_files_in_dir(dir_)

    #for fpath in cython_files:
    #    fname, ext = splitext(fpath)
    #    for libext in util_cplat.LIB_EXT_LIST:
    #        util_path.remove_file(fname + libext)
    #    util_path.remove_file(fname + '.c')


#def build_cython(cython_files):
#    """ doesn't work """
#    for fpath in cython_files:
#        util_dev.compile_cython(fpath)


# def translate_cyth():
#     import cyth
#     cyth.translate_all()


def get_numpy_include_dir():
    try:
        import numpy as np
        return np.get_include()
    except ImportError:
        return ''


def find_ext_modules(disable_warnings=True):
    from setuptools import Extension
    import utool
    from os.path import relpath
    cwd = os.getcwd()

    # CYTH      = 'cyth' in sys.argv
    BEXT      = 'bext' in sys.argv
    BUILD     = 'build' in sys.argv
    BUILD_EXT = 'build_ext' in sys.argv

    # if any([BEXT, CYTH]):
    #     translate_cyth()  # translate cyth before finding ext modules

    if not any([BEXT, BUILD, BUILD_EXT]):
        # dont find modules if they are not being built
        return []

    #pyx_list = utool.glob(cwd, '*_cython.pyx', recursive=True)
    pyx_list = utool.glob(cwd, '*.pyx', recursive=True)

    if disable_warnings:
        extra_compile_args = ['-Wno-format', '-Wno-unused-function']
    else:
        extra_compile_args = []

    ext_modules = []
    for pyx_abspath in pyx_list:
        pyx_relpath = relpath(pyx_abspath, cwd)
        pyx_modname, _ = splitext(pyx_relpath.replace('\\', '.').replace('/', '.'))
        print('[find_ext] Found Module:')
        print('   * pyx_modname = %r' % (pyx_modname,))
        print('   * pyx_relpath = %r' % (pyx_relpath,))
        extmod = Extension(pyx_modname, [pyx_relpath],
                           include_dirs=[get_numpy_include_dir()],
                           extra_compile_args=extra_compile_args)
        ext_modules.append(extmod)
    return ext_modules


def find_packages(recursive=True, maxdepth=None):
    """
    Finds all directories with an __init__.py file in them
    """
    import utool
    if utool.VERBOSE:
        print('[util_setup] find_packages(recursive=%r, maxdepth=%r)' % (recursive, maxdepth))
    from os.path import relpath
    cwd = os.getcwd()
    init_files = utool.glob(cwd, '__init__.py', recursive=recursive, maxdepth=maxdepth)
    package_paths = list(map(dirname, init_files))
    package_relpaths = [relpath(path, cwd) for path in package_paths]

    packages = []
    for path in package_relpaths:
        base = utool.dirsplit(path)[0]
        if exists(join(base, '__init__.py')):
            package = path.replace('/', '.').replace('\\', '.')
            packages.append(package)
    return packages


def get_cmdclass():
    """ DEPRICATE """
    try:
        from Cython.Distutils import build_ext
        cmdclass = {'build_ext': build_ext}
        return cmdclass
    except Exception as ex:
        print(ex)
        print('WARNING: Cython is not installed. This is only a problem if you are building C extensions')
        return {}


def parse_author():
    """ TODO: this function should parse setup.py or a module for
    the author variable
    """
    return 'Jon Crall'   # FIXME


def autogen_sphinx_apidoc():
    r"""
    autogen_sphinx_docs.py

    Ignore:
        C:\Python27\Scripts\autogen_sphinx_docs.py
        autogen_sphinx_docs.py

        pip uninstall sphinx
        pip install sphinx
        pip install sphinxcontrib-napoleon
        pip install sphinx --upgrade
        pip install sphinxcontrib-napoleon --upgrade

        cd C:\Python27\Scripts
        ls C:\Python27\Scripts

        python -c "import sphinx; print(sphinx.__version__)"

    CommandLine:
        python -m utool.util_setup --exec-autogen_sphinx_apidoc

    Example:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> from utool.util_setup import *  # NOQA
        >>> autogen_sphinx_apidoc()
    """
    # TODO: assert sphinx-apidoc exe is found
    # TODO: make find_exe work?
    import utool as ut

    def build_sphinx_apidoc_cmdstr():
        print('')
        print('if this fails try: sudo pip install sphinx')
        print('')
        apidoc = 'sphinx-apidoc'
        if ut.WIN32:
            winprefix = 'C:/Python27/Scripts/'
            sphinx_apidoc_exe = winprefix + apidoc + '.exe'
        else:
            sphinx_apidoc_exe = apidoc
        apidoc_argfmt_list = [
            sphinx_apidoc_exe,
            '--force',
            '--full',
            '--maxdepth="{maxdepth}"',
            '--doc-author="{author}"',
            '--doc-version="{doc_version}"',
            '--doc-release="{doc_release}"',
            '--output-dir="_doc"',
            #'--separate',  # Put documentation for each module on its own page
            '--private',  # Include "_private" modules
            '{pkgdir}',
        ]
        outputdir = '_doc'
        author = ut.parse_author()
        packages = ut.find_packages(maxdepth=1)
        assert len(packages) != 0, 'directory must contain at least one package'
        if len(packages) > 1:
            assert len(packages) == 1,\
                ('FIXME I dont know what to do with more than one root package: %r'
                 % (packages,))
        pkgdir = packages[0]
        version = ut.parse_package_for_version(pkgdir)
        modpath = dirname(ut.truepath(pkgdir))

        apidoc_fmtdict = {
            'author': author,
            'maxdepth': '8',
            'pkgdir': pkgdir,
            'doc_version': version,
            'doc_release': version,
            'outputdir': outputdir,
        }
        ut.assert_exists('setup.py')
        ut.ensuredir('_doc')
        apidoc_fmtstr = ' '.join(apidoc_argfmt_list)
        apidoc_cmdstr = apidoc_fmtstr.format(**apidoc_fmtdict)
        print('[util_setup] autogenerate sphinx docs for %r' % (pkgdir,))
        if ut.VERBOSE:
            print(ut.repr4(apidoc_fmtdict))
        return apidoc_cmdstr, modpath, outputdir

    def build_conf_replstr():
        #
        # Make custom edits to conf.py
        # FIXME:
        #ext_search_text = ut.unindent(
        #    r'''
        #    extensions = [
        #    [^\]]*
        #    ]
        #    ''')
        ext_search_text = r'extensions = \[[^/]*\]'
        # TODO: http://sphinx-doc.org/ext/math.html#module-sphinx.ext.pngmath
        #'sphinx.ext.mathjax',
        exclude_modules = []
        ext_repl_text = ut.codeblock(
            '''
            MOCK_MODULES = {exclude_modules}
            if len(MOCK_MODULES) > 0:
                import mock
                for mod_name in MOCK_MODULES:
                    sys.modules[mod_name] = mock.Mock()

            extensions = [
                'sphinx.ext.autodoc',
                'sphinx.ext.viewcode',
                # For LaTeX
                'sphinx.ext.pngmath',
                # For Google Sytle Docstrs
                # https://pypi.python.org/pypi/sphinxcontrib-napoleon
                'sphinxcontrib.napoleon',
                #'sphinx.ext.napoleon',
            ]
            '''
        ).format(exclude_modules=str(exclude_modules))
        #theme_search = 'html_theme = \'default\''
        theme_search = 'html_theme = \'[a-zA-Z_1-3]*\''
        theme_repl = ut.codeblock(
            '''
            import sphinx_rtd_theme
            html_theme = "sphinx_rtd_theme"
            html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
            ''')
        head_text = ut.codeblock(
            '''
            from sphinx.ext.autodoc import between
            import sphinx_rtd_theme
            import sys
            import os

            # Dont parse IBEIS args
            os.environ['IBIES_PARSE_ARGS'] = 'OFF'
            os.environ['UTOOL_AUTOGEN_SPHINX_RUNNING'] = 'ON'

            sys.path.append('{modpath}')
            sys.path.append(sys.path.insert(0, os.path.abspath("../")))

            autosummary_generate = True

            modindex_common_prefix = ['_']
            '''
        ).format(modpath=ut.truepath(modpath))
        tail_text = ut.codeblock(
            '''
            def setup(app):
                # Register a sphinx.ext.autodoc.between listener to ignore everything
                # between lines that contain the word IGNORE
                app.connect('autodoc-process-docstring', between('^.*IGNORE.*$', exclude=True))
                return app
            '''
        )
        return (ext_search_text, ext_repl_text, theme_search, theme_repl, head_text, tail_text)

    apidoc_cmdstr, modpath, outputdir = build_sphinx_apidoc_cmdstr()
    ext_search_text, ext_repl_text, theme_search, theme_repl, head_text, tail_text = build_conf_replstr()

    dry = ut.get_argflag('--dry')

    if not dry:
        # Execute sphinx-apidoc
        ut.cmd(apidoc_cmdstr, shell=True)
        # sphinx-apidoc outputs conf.py to <outputdir>, add custom commands
        #
        # Change dir to <outputdir>
        print('chdir' + outputdir)
        os.chdir(outputdir)
        conf_fname = 'conf.py'
        conf_text = ut.read_from(conf_fname)
        conf_text = conf_text.replace('import sys', 'import sys  # NOQA')
        conf_text = conf_text.replace('import os', 'import os  # NOQA')
        conf_text = ut.regex_replace(theme_search, theme_repl, conf_text)
        conf_text = ut.regex_replace(ext_search_text, ext_repl_text, conf_text)
        conf_text = head_text + '\n' + conf_text + tail_text
        ut.write_to(conf_fname, conf_text)
        # Make the documentation
        #if ut.LINUX:
        #    ut.cmd('make html', shell=True)
        #if ut.WIN32:
        #raw_input('waiting')
        if not ut.get_argflag('--nomake'):
            ut.cmd('make', 'html', shell=True)
    else:
        print(apidoc_cmdstr)
        print('cd ' + outputdir)
        print('manual edits of conf.py')
        print('make html')


def NOOP():
    pass


def presetup_commands(setup_fpath, kwargs):
    if VERBOSE:
        print('[setup] presetup_commands()')
    name = kwargs.get('name', '')
    # Parse args
    project_dirs     = kwargs.pop('project_dirs', None)
    chmod_patterns   = kwargs.pop('chmod_patterns', [])
    clutter_dirs     = kwargs.pop('clutter_dirs', None)
    clutter_patterns = kwargs.pop('clutter_patterns', [])
    build_command    = kwargs.pop('build_command', NOOP)
    # Augment patterns with builtin patterns
    chmod_patterns   += SETUP_PATTERNS.chmod_test if kwargs.pop('chmod_tests', True) else []
    clutter_patterns += SETUP_PATTERNS.clutter_pybuild if kwargs.pop('clean_pybuild', True) else []
    clutter_patterns += SETUP_PATTERNS.clutter_cyth if kwargs.pop('clean_pybuild', True) else []
    setup_fpath = util_path.truepath(setup_fpath)
    #
    setup_dir = dirname(setup_fpath)
    build_dir = join(setup_dir, 'build')
    os.chdir(setup_dir)  # change into setup directory
    assert_in_setup_repo(setup_fpath, name)

    # Augment clutter dirs
    if clutter_dirs is None:
        clutter_dirs = []
    clutter_dirs += [
        'build',
        'dist',
        name + '.egg-info',
        '__pycache__'
    ]

    if project_dirs is None:
        project_dirs = util_path.ls_moduledirs(setup_dir)
# Execute pre-setup commands based on argv
    #BEXT = 'bext' in sys.argv
    #BUILD_EXT = 'build_ext' in sys.argv
    #CYTH = 'cyth' in sys.argv

    #if CYTH:
    #    translate_cyth()

    for arg in iter(sys.argv[:]):
        #print(arg)
        # Clean clutter files
        if arg in ['clean']:
            clean(setup_dir, clutter_patterns, clutter_dirs)
            #sys.exit(0)
        if arg in ['build'] or (not exists(build_dir) and
                                'clean' not in sys.argv):
            if VERBOSE:
                print('[setup] Executing build command')
            try:
                build_command()
            except Exception as ex:
                printex(ex, 'Error calling buildcommand from cwd=%r\n' % os.getcwd())
                raise
        if arg in ['versions']:
            print('+ ---')
            print('Testing preqres package versions')
            install_requires = kwargs.get('install_requires', [])
            import pip
            print('Checking install_requires = [\n%s\n]' % '\n'.join(install_requires))
            pip.main(['show'] + [depline.split(' ')[0] for depline in install_requires])
            print('L ___ Done Version Testing')
        if arg in ['docs']:
            # FIXME
            autogen_sphinx_apidoc()
        # Build optimized files
        if arg in ['o', 'pyo']:
            build_pyo(project_dirs)
        # Cythonize files
        #if arg in ['cython']:
        #    build_cython(cython_files)
        # Chmod files
        if arg in ['chmod']:
            setup_chmod(setup_fpath, setup_dir, chmod_patterns)

    try:
        sys.argv.remove('docs')
        sys.argv.remove('cyth')
        #sys.argv.remove('--cyth-write')
    except ValueError:
        pass

    try:
        # SUPER HACK
        # aliases bext to build_ext --inplace
        sys.argv.remove('bext')
        sys.argv.append('build_ext')
        sys.argv.append('--inplace')
    except ValueError:
        pass


presetup = presetup_commands


# def parse_package_version(fpath='__init__.py', varname='__version__'):
#     """ Statically parse the version number from __init__.py """
#     # from os.path import dirname, join
#     import ast
#     # repo_dpath = dirname(__file__)
#     # file_fpath = join(repo_dpath, 'ibeis', '__init__.py')
#     # file_fpath = join(repo_dpath, 'ibeis', 'control', 'DB_SCHEMA_CURRENT.py')
#     with open(fpath) as file_:
#         sourcecode = file_.read()
#     pt = ast.parse(sourcecode)

#     class VersionVisitor(ast.NodeVisitor):
#         def visit_Assign(self, node):
#             for target in node.targets:
#                 if target.id == varname:
#                     self.version = node.value.s
#     visitor = VersionVisitor()
#     visitor.visit(pt)
#     return visitor.version


def parse_package_for_version(name):
    """
    Searches for a variable named __version__ in name's __init__.py file and
    returns the value.  This function parses the source text. It does not load
    the module.
    """
    from utool import util_regex
    init_fpath = join(name, '__init__.py')
    version_errmsg = textwrap.dedent(
        '''
        You must include a __version__ variable
        in %s\'s __init__.py file.
        Try something like:
        __version__ = '1.0.0.dev1' ''' % (name,))
    if not exists(init_fpath):
        raise AssertionError(version_errmsg)
    val_regex = util_regex.named_field('version', '[0-9a-zA-Z.]+')
    regexstr = '__version__ *= *[\'"]' + val_regex
    def parse_version(line):
        # Helper
        line = line.replace(' ', '').replace('\t', '')
        match_dict = util_regex.regex_parse(regexstr, line)
        if match_dict is not None:
            return match_dict['version']
    # Find the version  in the text of the source
    #version = 'UNKNOWN_VERSION'
    with open(init_fpath, 'r') as file_:
        for line in file_.readlines():
            if line.startswith('__version__'):
                version = parse_version(line)
                if version is not None:
                    return version
    raise AssertionError(version_errmsg)


def __infer_setup_kwargs(module, kwargs):
    """ Implicitly build kwargs based on standard info """
    # Get project name from the module
    #if 'name' not in kwargs:
    #    kwargs['name'] = module.__name__
    #else:
    #    raise AssertionError('must specify module name!')
    name = kwargs['name']
    # Our projects depend on utool
    #if kwargs['name'] != 'utool':
    #    install_requires = kwargs.get('install_requires', [])
    #    if 'utool' not in install_requires:
    #        install_requires.append('utool')
    #    kwargs['install_requires'] = install_requires

    packages = kwargs.get('packages', [])
    if name not in packages:
        packages.append(name)
        kwargs['packages'] = packages

    if 'version' not in kwargs:
        version = parse_package_for_version(name)
        kwargs['version'] = version

    # Parse version
    #if 'version' not in kwargs:
    #    if module is None:
    #        version_errmsg = 'You must include a version (preferably one that matches the __version__ variable in your modules init file'
    #        raise AssertionError(version_errmsg)
    #    else:
    # Parse license
    if 'license' not in kwargs:
        try:
            kwargs['license'] = read_license('LICENSE')
        except IOError:
            pass
    # Parse readme
    if 'long_description' not in kwargs:
        kwargs['long_description'] = parse_readme()


def parse_readme(readmefile='README.md'):
    # DEPRICATE
    return util_io.read_from(readmefile, verbose=False, strict=False)


def setuptools_setup(setup_fpath=None, module=None, **kwargs):
    # TODO: Learn this better
    # https://docs.python.org/3.1/distutils/apiref.html
    # https://pythonhosted.org/an_example_pypi_project/setuptools.html
    # https://docs.python.org/2/distutils/setupscript.html https://docs.python.org/2/distutils/setupscript.html
    # Useful documentation: http://bashelton.com/2009/04/setuptools-tutorial/#setup.py-package_dir
    """
    Arguments which can be passed to setuptools::

        ============       =====            ===========
        Install-Data       Value            Description
        ------------       -----            -----------
        *packages          strlist          a list of packages modules to be distributed
        py_modules         strlist          a list of singlefile modules to be distributed
        scripts            strlist          a list of standalone scripts to build and install
        *install_requires  list             e.g: ['distribute == 0.7.3', 'numpy', 'matplotlib']
        data_files         strlist          a list of data files to install
        zip_safe           bool             install efficiently installed as a zipped module?
        namespace_packages list             packages without meaningful __init__.py's
        package_dir        dict             keys are packagenames ('' is the root)
        package_data       dict             keys are foldernames, values are a list of globstrs
        *entry_pionts      dict             installs a script {'console_scripts': ['script_name_to_install = entry_module:entry_function']}

        ============       =====            ===========
        Meta-Data          Value            Description
        ------------       -----            -----------
        name               short string     ('name of the package')
        version            short string     ('version of this release')
        author             short string     ('package authors name')
        author_email       email address    ('email address of the package author')
        maintainer         short string     ('package maintainers name')
        maintainer_email   email address    ('email address of the package maintainer')
        url                URL              ('home page for the package')
        description        short string     ('short, summary description of the package')
        long_description   long string      ('longer description of the package')
        download_url       URL              ('location where the package may be downloaded')
        classifiers        list of strings  ('a list of classifiers')
        platforms          list of strings  ('a list of platforms')
        license            short string     ('license for the package')
    """
    from utool.util_inject import inject_colored_exceptions
    inject_colored_exceptions()  # Fluffly, but nice
    if VERBOSE:
        print(util_str.repr4(kwargs))
    __infer_setup_kwargs(module, kwargs)
    presetup_commands(setup_fpath, kwargs)
    if VERBOSE:
        print(util_str.repr4(kwargs))
    return kwargs


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_setup
        python -m utool.util_setup --allexamples
        python -m utool.util_setup --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



### __init__.py ###
# flake8: noqa
"""
Need to hack to make work for internal modules and reload subs properly

Regen Command:
    cd /home/joncrall/code/utool/utool/_internal
    makeinit.py -x win32_send_keys
"""
# autogenerated __init__.py for: '/home/joncrall/code/utool/utool/_internal'



# -*- coding: utf-8 -*-
"""
UTool - Useful Utility Tools
   Your friendly neighborhood utility tools

TODO: INSERT APACHE VERSION 2.0 LICENCE IN ALL FILES
(Although it should be implied that the entire module and repo is released under
that licence.)

pip install git+https://github.com/Erotemic/utool.git@next
"""
# Utool is released under the Apache License Version 2.0

# flake8: noqa
# We hope to support python3
from __future__ import absolute_import, division, print_function, unicode_literals
import sys
import textwrap

__version__ = '2.1.5'

__DYNAMIC__ = True
if __DYNAMIC__:
    __DYNAMIC__ = '--nodyn' not in sys.argv
else:
    __DYNAMIC__ = '--dyn' in sys.argv
# THESE COMMANDS WILL WRITE THE IMPORT FILE
"""
python -c "import utool" --dump-utool-init
python -c "import utool" --print-utool-init --dyn
python -c "import utool" --update-utool-init --dyn
"""

__VERYVERBOSE__ = '--veryverbose' in sys.argv or '--very-verbose' in sys.argv
__PRINT_INJECT_ORDER__ = __VERYVERBOSE__ or '--print-inject-order' in sys.argv

# HAVE TO HACK THIS IN FOR UTOOL.__INIT__ ONLY
# OTHER MODULE CAN USE NOINJECT
if __PRINT_INJECT_ORDER__:
    # TODO METAUTIL
    from six.moves import builtins
    from utool._internal import meta_util_dbg
    callername = meta_util_dbg.get_caller_name(N=1, strict=False)
    fmtdict = dict(callername=callername, modname='utool.__init__')
    msg = '[util_inject] {modname} is imported by {callername}'.format(**fmtdict)
    builtins.print(msg)


IMPORT_TUPLES = [
    ('_internal',      None),
    ('util_alg',       ['cartesian', 'almost_eq',]),
    ('util_aliases',   ['ddict' ,'odict']),
    ('util_arg',       ['get_argval', 'get_argflag', 'argv_flag_dec', 'QUIET',
                        'VERBOSE']),
    ('util_assert',    None),
    ('util_autogen',   None),
    ('util_cache',     ['global_cache_read', 'global_cache_write']),
    ('util_cplat',     ['cmd', 'view_directory',]),
    ('util_class',     None),
    ('util_const',     None),
    ('util_csv',       None),
    ('util_config',    None),
    ('util_dbg',       ['execstr_dict', 'save_testdata', 'load_testdata',
                        'get_caller_name', 'import_testdata', 'embed',
                        'quitflag', 'inIPython', 'printvar2', 'all_rrr']),
    ('util_dev',       ['get_stats_str', 'get_stats', 'myprint',
                        'get_object_nbytes']),
    ('util_decor',     ['ignores_exc_tb', 'indent_func', 'accepts_numpy',
                        'accepts_scalar_input_vector_output',
                        'accepts_scalar_input']),
    ('util_dict',      None),
    ('util_func',      None),
    ('util_grabdata',  None),
    ('util_gridsearch',  None),
    ('util_git',       None),
    ('util_latex',     None),
    ('util_graph',   None),
    ('util_hash',      ['hashstr_arr', 'hashstr',]),
    ('util_import',    None),
    ('util_inject',    None),
    ('util_io',        None),
    ('util_iter',      ['iflatten', 'ichunks',]),
    ('util_inspect',   None),
    ('util_ipynb',     None),
    ('util_logging',   None),
    ('util_list',      ['alloc_lists', 'list_index', 'npfind', 'index_of',
                        'flatten']),
    ('util_num',       None),
    ('util_numpy',       None),
    ('util_path',      ['checkpath', 'ensuredir', 'assertpath', 'truepath',
                        'list_images', 'copy', 'glob', 'grep']),
    ('util_print',     ['horiz_print', 'printshape', 'Indenter']),
    ('util_progress',  ['progress_func']),
    ('util_profile',   None),
    ('util_project',   None),
    ('util_parallel',  None),
    ('util_resources', ['print_resource_usage']),
    ('util_str',       ['byte_str2', 'horiz_string', 'theta_str']),
    ('util_sysreq',    None),
    ('util_sqlite',    None),
    ('util_setup',     'presetup'),
    ('util_set',       None),
    ('util_regex',     ['regex_search']),
    ('util_time',      ['tic', 'toc', 'Timer']),
    ('util_tags',      None),
    ('util_type',      None),
    ('util_tests',     None),
    ('util_web',     None),
    ('DynamicStruct',  ['DynStruct']),
    ('Preferences',    ['Pref']),
    ]


DOELSE = False

if __DYNAMIC__:
    # TODO: import all utool external prereqs. Then the imports will not import
    # anything that has already in a toplevel namespace
    # COMMENTED OUT FOR FROZEN __INIT__
    # Dynamically import listed util libraries and their members.
    from utool._internal import util_importer
    # FIXME: this might actually work with rrrr, but things arent being
    # reimported because they are already in the modules list
    import_execstr = util_importer.dynamic_import(__name__, IMPORT_TUPLES)
    exec(import_execstr)
    DOELSE = False
else:
    # Do the nonexec import (can force it to happen no matter what if alwyas set
    # to True)
    DOELSE = True

if DOELSE:
    # <AUTOGEN_INIT>

    from utool import _internal
    from utool import util_alg
    from utool import util_aliases
    from utool import util_arg
    from utool import util_assert
    from utool import util_autogen
    from utool import util_cache
    from utool import util_cplat
    from utool import util_class
    from utool import util_const
    from utool import util_csv
    from utool import util_config
    from utool import util_dbg
    from utool import util_dev
    from utool import util_decor
    from utool import util_dict
    from utool import util_func
    from utool import util_grabdata
    from utool import util_gridsearch
    from utool import util_git
    from utool import util_latex
    from utool import util_graph
    from utool import util_hash
    from utool import util_import
    from utool import util_inject
    from utool import util_io
    from utool import util_iter
    from utool import util_inspect
    from utool import util_ipynb
    from utool import util_logging
    from utool import util_list
    from utool import util_num
    from utool import util_numpy
    from utool import util_path
    from utool import util_print
    from utool import util_progress
    from utool import util_profile
    from utool import util_project
    from utool import util_parallel
    from utool import util_resources
    from utool import util_str
    from utool import util_sysreq
    from utool import util_sqlite
    from utool import util_setup
    from utool import util_set
    from utool import util_regex
    from utool import util_time
    from utool import util_tags
    from utool import util_type
    from utool import util_tests
    from utool import util_web
    from utool import DynamicStruct
    from utool import Preferences

    from utool.util_alg import (FOOT_PER_MILE, HAVE_NUMPY, HAVE_SCIPY,
                                KM_PER_MILE, MM_PER_INCH, PHI, PHI_A, PHI_B,
                                TAU, absdiff, almost_allsame, almost_eq,
                                apply_grouping, bayes_rule, choose,
                                colwise_diag_idxs, compare_groups, cumsum,
                                defaultdict, deg_to_rad, diagonalized_iter,
                                edit_distance, enumerate_primes,
                                euclidean_dist, expensive_task_gen, factors,
                                fibonacci, fibonacci_approx,
                                fibonacci_iterative, fibonacci_recursive,
                                find_group_consistencies,
                                find_group_differences,
                                flatten_membership_mapping, generate_primes,
                                get_nth_bell_number, get_nth_prime,
                                get_nth_prime_bruteforce, get_phi,
                                get_phi_ratio1, get_prime_index,
                                greedy_max_inden_setcover, group_indices,
                                grouping_delta, grouping_delta_stats,
                                iapply_grouping, inbounds, is_prime, item_hist,
                                knapsack, knapsack_greedy, knapsack_ilp,
                                knapsack_iterative, knapsack_iterative_int,
                                knapsack_iterative_numpy, knapsack_recursive,
                                longest_common_substring,
                                max_size_max_distance_subset,
                                maximin_distance_subset1d,
                                maximum_distance_subset,
                                negative_minclamp_inplace, norm_zero_one,
                                normalize, num_partitions, number_of_decimals,
                                prod, product_nonsame, product_nonsame_self,
                                rad_to_deg, safe_div, safe_pdist, self_prodx,
                                setcover_greedy, setcover_ilp, solve_boolexpr,
                                square_pdist, standardize_boolexpr,
                                triangular_number, ungroup, ungroup_gen,
                                ungroup_unique, unixtime_hourdiff,
                                upper_diag_self_prodx, xywh_to_tlbr,)
    from utool.util_aliases import (OrderedDict, combinations, ddict, icomb,
                                    iprod, namedtuple, odict, partial,
                                    product,)
    from utool.util_arg import (ArgumentParser2, DEBUG, DEBUG2, NOT_QUIET,
                                NO_ASSERTS, QUIET, REPORT, SAFE, SILENT,
                                STRICT, SUPER_STRICT, TRACE, USE_ASSERT,
                                VERBOSE, VERBOSE_ARGPARSE, VERYVERBOSE,
                                argflag, argparse_dict, argv_flag_dec,
                                argv_flag_dec_true, argval, aug_sysargv,
                                autogen_argparse2, autogen_argparse_block,
                                get_arg, get_arg_dict, get_argflag,
                                get_argv_tail, get_argval, get_cmdline_varargs,
                                get_dict_vals_from_commandline, get_flag,
                                get_fpath_args, get_module_verbosity_flags,
                                get_varargs, get_verbflag, make_argparse2,
                                parse_arglist_hack, parse_cfgstr_list,
                                parse_dict_from_argv, reset_argrecord,
                                switch_sanataize,)
    from utool.util_assert import (assert_all_eq, assert_all_in,
                                   assert_all_not_None, assert_almost_eq,
                                   assert_eq, assert_eq_len, assert_inbounds,
                                   assert_lessthan, assert_lists_eq,
                                   assert_raises, assert_same_len,
                                   assert_scalar_list, assert_unflat_level,
                                   assert_unique, get_first_None_position,
                                   lists_eq,)
    from utool.util_autogen import (PythonStatement, auto_docstr,
                                    autofix_codeblock, deque,
                                    dump_autogen_code,
                                    find_modname_in_pythonpath,
                                    is_modname_in_pythonpath,
                                    load_func_from_module, make_args_docstr,
                                    make_cmdline_docstr, make_default_docstr,
                                    make_default_module_maintest,
                                    make_docstr_block, make_example_docstr,
                                    make_returns_or_yeilds_docstr, makeinit,
                                    print_auto_docstr,
                                    remove_codeblock_syntax_sentinals,
                                    write_modscript_alias,)
    from utool.util_cache import (Cachable, CacheMissException, Cacher,
                                  GlobalShelfContext, KeyedDefaultDict,
                                  LRUDict, LazyDict, LazyList, ShelfCacher,
                                  USE_CACHE, VERBOSE_CACHE, cached_func,
                                  cachestr_repr, chain, consensed_cfgstr,
                                  delete_global_cache, from_json,
                                  get_cfgstr_from_args, get_default_appname,
                                  get_func_result_cachekey,
                                  get_global_cache_dir, get_global_shelf_fpath,
                                  get_lru_cache, global_cache_dump,
                                  global_cache_read, global_cache_write,
                                  load_cache, make_utool_json_encoder,
                                  save_cache, shelf_open, text_dict_read,
                                  text_dict_write, time_different_diskstores,
                                  to_json, tryload_cache, tryload_cache_list,
                                  tryload_cache_list_with_compute,
                                  view_global_cache_dir,)
    from utool.util_cplat import (COMPUTER_NAME, DARWIN, EXIT_FAILURE,
                                  EXIT_SUCCESS, HAVE_PATHLIB, LIB_DICT,
                                  LIB_EXT_LIST, LINUX, OS_TYPE, PYLIB_DICT,
                                  UNIX, WIN32, assert_installed_debian,
                                  change_term_title, check_installed_debian,
                                  chmod, chmod_add_executable, cmd, cmd2,
                                  editfile, ensure_app_cache_dir,
                                  ensure_app_resource_dir, get_app_cache_dir,
                                  get_app_resource_dir, get_computer_name,
                                  get_dir_diskspaces, get_disk_space,
                                  get_dynamic_lib_globstrs,
                                  get_dynlib_dependencies, get_dynlib_exports,
                                  get_file_info, get_file_nBytes,
                                  get_file_nBytes_str, get_flops,
                                  get_free_diskbytes, get_install_dirs,
                                  get_lib_ext, get_path_dirs,
                                  get_plat_specifier, get_pylib_ext,
                                  get_python_dynlib, get_resource_dir,
                                  get_system_python_library,
                                  get_total_diskbytes, get_user_name,
                                  geteditor, getroot, in_pyinstaller_package,
                                  ipython_paste, is64bit_python,
                                  is_file_executable, is_file_writable,
                                  ls_libs, pip_install, platform_cache_dir,
                                  print_dir_diskspace, print_path,
                                  print_system_users, python_executable,
                                  quote_single_command, run_realtime_process,
                                  search_env_paths, send_keyboard_input,
                                  set_process_title, shell,
                                  spawn_delayed_ipython_paste, startfile,
                                  truepath, unixpath, unload_module, vd,
                                  view_directory, view_file_in_directory,)
    from utool.util_class import (HashComparable, HashComparable2,
                                  HashComparableMetaclass, KwargsWrapper,
                                  QUIET_CLASS, ReloadingMetaclass,
                                  VERBOSE_CLASS,
                                  autogen_explicit_injectable_metaclass,
                                  autogen_import_list, compare_instance,
                                  decorate_class_method, decorate_postinject,
                                  get_classname, get_comparison_methods,
                                  get_comparison_operators, get_funcglobals,
                                  get_funcname, get_injected_modules,
                                  get_method_func, inject_all_external_modules,
                                  inject_func_as_method,
                                  inject_func_as_property,
                                  inject_func_as_unbound_method,
                                  inject_instance, makeForwardingMetaclass,
                                  make_class_method_decorator,
                                  make_class_postinject_decorator,
                                  postinject_instance, reload_class,
                                  reload_class_methods,
                                  reload_injected_modules, reloadable_class,
                                  reloading_meta_metaclass_factory,
                                  remove_private_obfuscation,
                                  test_reloading_metaclass,)
    from utool.util_const import (NoParam,)
    from utool.util_csv import (CSV, make_csv_table, make_standard_csv,
                                numpy_to_csv, read_csv,)
    from utool.util_config import (get_default_global_config,
                                   get_default_repo_config, read_repo_config,
                                   write_default_repo_config,)
    from utool.util_dbg import (COLORED_EXCEPTIONS, EmbedOnException, FORCE_TB,
                                RAISE_ALL, TB, breakpoint, debug_exception,
                                debug_hstack, debug_list, debug_npstack,
                                debug_vstack, embed, embed2,
                                embed_on_exception_context, eoxc,
                                execstr_attr_list, execstr_dict, execstr_func,
                                execstr_parent_locals, explore_module,
                                explore_stack, fix_embed_globals, fmtlocals,
                                formatex, get_caller_lineno,
                                get_caller_modname, get_caller_name,
                                get_caller_prefix, get_current_stack_depth,
                                get_localvar_from_stack, get_parent_frame,
                                get_reprs, get_stack_frame, get_var_from_stack,
                                get_varname_from_locals,
                                get_varname_from_stack, get_varstr,
                                get_varval_from_locals, haveIPython,
                                import_testdata, inIPython,
                                in_jupyter_notebook, ipython_execstr,
                                is_valid_varname, load_testdata,
                                module_functions, parse_locals_keylist,
                                print_frame, print_keys, print_traceback,
                                printex, printvar, printvar2,
                                public_attributes, qflag, quasiquote, quit,
                                quitflag, save_testdata,
                                search_stack_for_localvar,
                                search_stack_for_var, split, super_print,
                                varname_regex,)
    from utool.util_dev import (ClassAttrDictProxy, ColumnLists,
                                DEVELOPER_MODE, DictLike_old, ENABLE_MEMTRACK,
                                INDEXABLE_TYPES, InteractiveIter,
                                InteractivePrompt, LIVE_INTERACTIVE_ITER,
                                MemoryTracker, NamedPartial, NiceRepr,
                                PriorityQueue, STAT_KEY_ORDER, Shortlist,
                                USER_MODE, are_you_sure, autopep8_diff,
                                compile_cython, copy_text_to_clipboard,
                                delayed_retry_gen, dev_ipython_copypaster,
                                disable_garbage_collection,
                                enable_garbage_collection, ensure_str_list,
                                ensureqt, exec_argparse_funckw, exec_funckw,
                                execstr_funckw, extract_timeit_setup, find_exe,
                                fix_super_reload, focusvim, garbage_collect,
                                get_clipboard, get_cython_exe,
                                get_dev_paste_code, get_jagged_stats,
                                get_nonconflicting_path_old,
                                get_nonconflicting_string, get_object_base,
                                get_object_nbytes, get_object_size_str,
                                get_overlaps, get_partial_func_name,
                                get_setdiff_info, get_setdiff_items,
                                get_statdict, get_stats, get_stats_str,
                                get_submodules_from_dpath, grace_period,
                                init_catch_ctrl_c, input_timeout, instancelist,
                                inverable_group_multi_list,
                                inverable_unique_two_lists, ipcopydev,
                                is_developer, iup, make_at_least_n_items_valid,
                                make_call_graph, make_object_graph,
                                memory_dump, numpy_list_num_bits,
                                overrideable_partial, pandas_reorder,
                                print_object_size, print_object_size_tree,
                                pylab_qt4, report_memsize, reset_catch_ctrl_c,
                                search_module, search_utool, set_clipboard,
                                set_overlap_items, set_overlaps, stats_dict,
                                strip_line_comments, super2, timeit_compare,
                                timeit_grid, tuples_to_unique_scalars,
                                uninvert_unique_two_lists,
                                user_cmdline_prompt,)
    from utool.util_decor import (IGNORE_TRACEBACK, NOINDENT_DECOR,
                                  ONEX_REPORT_INPUT, PROFILING, SIG_PRESERVE,
                                  UNIQUE_NUMPY, accepts_numpy,
                                  accepts_scalar_input, accepts_scalar_input2,
                                  accepts_scalar_input_vector_output,
                                  apply_docstr, classproperty,
                                  debug_function_exceptions, dummy_args_decor,
                                  getter_1to1, getter_1toM, ignores_exc_tb,
                                  indent_func, interested, lazyfunc, memoize,
                                  memoize_nonzero, memoize_single,
                                  memoize_zero, on_exception_report_input,
                                  preserve_sig, show_return_value,
                                  test_ignore_exec_traceback, time_func,
                                  tracefunc, tracefunc_xml,)
    from utool.util_dict import (AutoDict, AutoOrderedDict, AutoVivification,
                                 DefaultValueDict, DictLike,
                                 OrderedAutoVivification,
                                 all_dict_combinations,
                                 all_dict_combinations_lbls,
                                 all_dict_combinations_ordered,
                                 assert_keys_are_subset, augdict,
                                 build_conflict_dict, count_dict_vals,
                                 delete_dict_keys, delete_keys, depth_atleast,
                                 dict_accum, dict_assign, dict_filter_nones,
                                 dict_find_keys, dict_find_other_sameval_keys,
                                 dict_hist, dict_hist_cumsum,
                                 dict_intersection, dict_isect,
                                 dict_isect_combine, dict_keysubset,
                                 dict_set_column, dict_setdiff, dict_stack,
                                 dict_stack2, dict_subset, dict_take,
                                 dict_take_asnametup, dict_take_column,
                                 dict_take_gen, dict_take_list, dict_take_pop,
                                 dict_to_keyvals, dict_union, dict_union2,
                                 dict_union3, dict_union_combine,
                                 dict_update_newkeys, dict_where_len0,
                                 dictinfo, dzip, flatten_dict_items,
                                 flatten_dict_vals, get_dict_column,
                                 get_dict_hashid, group_items, group_pairs,
                                 groupby_attr, groupby_tags, hashdict,
                                 hierarchical_group_items,
                                 hierarchical_map_vals, hmap_vals,
                                 iflatten_dict_values, invert_dict, is_dicteq,
                                 iter_all_dict_combinations_ordered,
                                 iteritems_sorted, keys_sorted_by_value,
                                 map_dict_keys, map_dict_vals, map_keys,
                                 map_vals, merge_dicts, move_odict_item,
                                 order_dict_by, range_hist, sort_dict,
                                 update_dict, update_existing,)
    from utool.util_func import (compose_functions, identity,)
    from utool.util_grabdata import (BadZipfile, HASH_DICT, TESTIMG_URL_DICT,
                                     archive_files, clean_dropbox_link,
                                     clear_test_img_cache, download_url,
                                     experiment_download_multiple_urls,
                                     geo_locate, get_file_local_hash,
                                     get_prefered_browser,
                                     get_valid_test_imgkeys,
                                     grab_file_remote_hash, grab_file_url,
                                     grab_s3_contents,
                                     grab_selenium_chromedriver,
                                     grab_selenium_driver, grab_test_imgpath,
                                     grab_zipped_url, list_remote,
                                     open_url_in_browser, read_s3_contents,
                                     rsync, s3_dict_encode_to_str,
                                     s3_str_decode_to_dict, scp_pull,
                                     split_archive_ext, unarchive_file,
                                     untar_file, unzip_file, url_read,
                                     url_read_text,)
    from utool.util_gridsearch import (CountstrParser, DimensionBasis,
                                       GridSearch, INTERNAL_CFGKEYS,
                                       NAMEVARSEP, Nesting, ParamInfo,
                                       ParamInfoBool, ParamInfoList,
                                       constrain_cfgdict_list,
                                       customize_base_cfg, get_cfg_lbl,
                                       get_cfgdict_lbl_list_subset,
                                       get_cfgdict_list_subset,
                                       get_nonvaried_cfg_lbls,
                                       get_varied_cfg_lbls,
                                       grid_search_generator, gridsearch_timer,
                                       interact_gridsearch_result_images,
                                       lookup_base_cfg_list, make_cfglbls,
                                       make_constrained_cfg_and_lbl_list,
                                       noexpand_parse_cfgstrs, parse_argv_cfg,
                                       parse_cfgstr3, parse_cfgstr_list2,
                                       parse_cfgstr_name_options,
                                       parse_nestings, parse_nestings2,
                                       partition_varied_cfg_list,
                                       recombine_nestings,
                                       testdata_grid_search,)
    from utool.util_git import (Repo, RepoManager, git_sequence_editor_squash,
                                isdir, std_build_command,)
    from utool.util_latex import (compile_latex_text, compress_pdf,
                                  convert_pdf_to_image, ensure_colvec,
                                  ensure_rowvec, escape_latex,
                                  find_ghostscript_exe, get_latex_figure_str,
                                  get_latex_figure_str2, is_substr,
                                  latex_get_stats, latex_multicolumn,
                                  latex_multirow, latex_newcommand,
                                  latex_sanitize_command_name, latex_scalar,
                                  long_substr, make_full_document,
                                  make_score_tabular, make_stats_tabular,
                                  render_latex, render_latex_text, replace_all,
                                  tabular_join,)
    from utool.util_graph import (all_multi_paths, approx_min_num_components,
                                  bfs_conditional, bfs_multi_edges,
                                  color_nodes, dag_longest_path,
                                  dfs_conditional, dict_depth,
                                  edges_to_adjacency_list, get_allkeys,
                                  get_graph_bounding_box, get_levels,
                                  graph_info, greedy_mincost_diameter_augment,
                                  longest_levels, mincost_diameter_augment,
                                  nx_all_nodes_between,
                                  nx_all_simple_edge_paths,
                                  nx_common_ancestors, nx_common_descendants,
                                  nx_contracted_nodes, nx_dag_node_rank,
                                  nx_delete_None_edge_attr,
                                  nx_delete_None_node_attr,
                                  nx_delete_edge_attr, nx_delete_node_attr,
                                  nx_edges, nx_edges_between,
                                  nx_ensure_agraph_color, nx_from_adj_dict,
                                  nx_from_matrix, nx_from_node_edge,
                                  nx_gen_edge_attrs, nx_gen_edge_values,
                                  nx_gen_node_attrs, nx_gen_node_values,
                                  nx_get_default_edge_attributes,
                                  nx_get_default_node_attributes,
                                  nx_make_adj_matrix, nx_mincut_edges_weighted,
                                  nx_minimum_weight_component, nx_node_dict,
                                  nx_set_default_edge_attributes,
                                  nx_set_default_node_attributes,
                                  nx_sink_nodes, nx_source_nodes,
                                  nx_to_adj_dict, nx_topsort_nodes,
                                  nx_topsort_rank, nx_transitive_reduction,
                                  paths_to_root, reverse_path,
                                  reverse_path_edges, shortest_levels,
                                  simplify_graph, stack_graphs,
                                  subgraph_from_edges, testdata_graph,
                                  translate_graph, translate_graph_to_origin,
                                  traverse_path, weighted_diamter,)
    from utool.util_hash import (ALPHABET, ALPHABET_16, ALPHABET_27,
                                 ALPHABET_41, BIGBASE, DictProxyType, HASH_LEN,
                                 HASH_LEN2, SEP_BYTE, SEP_STR, augment_uuid, b,
                                 combine_hashes, combine_uuids,
                                 convert_bytes_to_bigbase,
                                 convert_hexstr_to_bigbase, digest_data,
                                 freeze_hash_bytes, get_file_hash,
                                 get_file_uuid, get_zero_uuid, hash_data,
                                 hashable_to_uuid, hashid_arr, hashstr,
                                 hashstr27, hashstr_arr, hashstr_arr27,
                                 hashstr_md5, hashstr_sha1, image_uuid,
                                 make_hash, random_nonce, random_uuid,
                                 stringlike, write_hash_file,
                                 write_hash_file_for_path,)
    from utool.util_import import (check_module_installed,
                                   get_modpath_from_modname, import_modname,
                                   import_module_from_fpath, import_star,
                                   import_star_execstr, package_contents,
                                   possible_import_patterns, tryimport,)
    from utool.util_inject import (ARGV_DEBUG_FLAGS, DEBUG_PRINT,
                                   DEBUG_PRINT_N, DEBUG_SLOW_IMPORT,
                                   DUMMYPROF_FUNC, EXIT_ON_INJECT_MODNAME,
                                   PRINT_INJECT_ORDER, PROFILE_FUNC,
                                   PROF_FUNC_PAT_LIST, PROF_MOD_PAT_LIST,
                                   TIMERPROF_FUNC, argv,
                                   colored_pygments_excepthook, inject,
                                   inject2, inject_colored_exceptions,
                                   inject_print_functions, inject_python_code,
                                   inject_python_code2, make_module_print_func,
                                   make_module_profile_func,
                                   make_module_reload_func,
                                   make_module_write_func, memprof, noinject,
                                   reload_module,
                                   split_python_text_into_lines,)
    from utool.util_io import (HAS_H5PY, HAS_NUMPY, HAVE_LOCKFILE, load_cPkl,
                               load_data, load_hdf5, load_json, load_numpy,
                               load_pytables, load_text, lock_and_load_cPkl,
                               lock_and_save_cPkl, read_from, read_lines_from,
                               readfrom, save_cPkl, save_data, save_hdf5,
                               save_json, save_numpy, save_pytables, save_text,
                               try_decode, write_to, writeto,)
    from utool.util_iter import (and_iters, ensure_iterable,
                                 evaluate_generator, ichunk_slices, ichunks,
                                 ichunks_cycle, ichunks_list, ichunks_noborder,
                                 ichunks_replicate, ifilter_Nones,
                                 ifilter_items, ifilterfalse_items, iflatten,
                                 iget_list_column, iget_list_column_slice,
                                 interleave, isiterable, isscalar,
                                 itake_column, iter_compress, iter_multichunks,
                                 iter_window, itertwo, next_counter,
                                 random_combinations, random_product,
                                 wrap_iterable,)
    from utool.util_inspect import (BaronWraper, KWReg, LIB_PATH,
                                    VERBOSE_INSPECT, VERYVERB_INSPECT,
                                    argparse_funckw, check_module_usage,
                                    check_static_member_vars, dummy_func,
                                    exec_func_doctest, exec_func_src,
                                    exec_func_src2, execstr_func_doctest,
                                    filter_valid_kwargs,
                                    find_funcs_called_with_kwargs,
                                    get_argnames, get_dev_hints, get_docstr,
                                    get_func_argspec, get_func_docblocks,
                                    get_func_kwargs, get_func_modname,
                                    get_func_sourcecode, get_funcdoc,
                                    get_funcfpath, get_funckw,
                                    get_funcnames_from_modpath,
                                    get_instance_attrnames,
                                    get_internal_call_graph, get_kwargs,
                                    get_kwdefaults, get_kwdefaults2,
                                    get_module_from_class,
                                    get_module_owned_functions,
                                    get_object_methods, get_unbound_args,
                                    help_members,
                                    infer_arg_types_and_descriptions,
                                    infer_function_info, inherit_kwargs,
                                    is_bateries_included, is_defined_by_module,
                                    is_defined_by_module2, is_valid_python,
                                    iter_module_doctestable,
                                    list_class_funcnames,
                                    list_global_funcnames,
                                    lookup_attribute_chain,
                                    parse_func_kwarg_keys,
                                    parse_function_names, parse_import_names,
                                    parse_kwarg_keys, parse_return_type,
                                    prettyprint_parsetree,
                                    recursive_parse_kwargs, set_funcdoc,
                                    set_funcname,
                                    special_parse_process_python_code,
                                    zzz_profiled_is_no, zzz_profiled_is_yes,)
    from utool.util_ipynb import (IPYNBCell, code_cell, export_notebook,
                                  format_cells, make_autogen_str,
                                  make_notebook, markdown_cell,
                                  normalize_cells, repr_single_for_md,
                                  run_ipython_notebook,)
    from utool.util_logging import (CustomStreamHandler, LOGGING_VERBOSE,
                                    PRINT_ALL_CALLERS, add_logging_handler,
                                    debug_logging_iostreams, ensure_logging,
                                    get_current_log_fpath,
                                    get_current_log_text, get_log_fpath,
                                    get_logging_dir, get_shelves_dir,
                                    get_utool_logger, is_logging,
                                    logdir_cacheid, start_logging,
                                    stop_logging, testlogprog,)
    from utool.util_list import (accumulate, alloc_lists, alloc_nones, allsame,
                                 and_lists, argmax, argmin, argsort, argsort2,
                                 aslist, broadcast_zip, bzip, ceil, compress,
                                 debug_consec_list, debug_duplicate_items,
                                 delete_items_by_index, delete_list_items,
                                 depth, depth_profile, duplicates_exist, emap,
                                 ensure_list_size, equal, estarmap, ezip,
                                 filter_Nones, filter_items, filter_startswith,
                                 filterfalse_items, find_duplicate_items,
                                 find_list_indexes, find_nonconsec_values,
                                 flag_None_items, flag_not_None_items,
                                 flag_percentile_parts, flag_unique_items,
                                 flat_unique, flatten, floor, get_dirty_items,
                                 get_list_column, get_list_column_slice,
                                 group_consecutives, group_consecutives_numpy,
                                 iflag_unique_items, index_complement,
                                 index_to_boolmask, insert_values,
                                 intersect_ordered, invertible_flatten1,
                                 invertible_flatten2,
                                 invertible_flatten2_numpy,
                                 invertible_total_flatten, is_subset,
                                 is_subset_of_any, is_superset, isdisjoint,
                                 isect, isect_indices, isetdiff_flags,
                                 issorted, issubset, issuperset, isunique,
                                 length_hint, listT, list_alignment,
                                 list_all_eq_to, list_argmax, list_argmaxima,
                                 list_argsort, list_compress,
                                 list_compresstake, list_cover,
                                 list_deep_types, list_depth, list_getattr,
                                 list_intersection, list_inverse_take,
                                 list_isdisjoint, list_issubset,
                                 list_issuperset, list_replace, list_reshape,
                                 list_roll, list_set_equal, list_strip,
                                 list_take, list_transpose, list_type,
                                 list_type_profile, list_union, list_where,
                                 list_zipcompress, list_zipflatten,
                                 list_ziptake, listclip, listfind, lmap,
                                 lstarmap, lzip, make_index_lookup,
                                 make_sortby_func, maplen, multi_replace,
                                 none_take, not_list, or_lists,
                                 partial_imap_1to1, partial_order,
                                 print_duplicate_map, priority_argsort,
                                 priority_sort, rebase_labels,
                                 recursive_replace, replace_nones,
                                 safe_listget, safe_slice, safeapply, safelen,
                                 sample_lists, sample_zip, scalar_input_map,
                                 search_list, setdiff, setdiff_flags,
                                 setdiff_ordered, setintersect,
                                 setintersect_ordered, snapped_slice, sortedby,
                                 sortedby2, strided_sample, take,
                                 take_around_percentile, take_column,
                                 take_complement, take_percentile,
                                 take_percentile_parts, total_flatten,
                                 total_unflatten, type_profile, type_profile2,
                                 type_sequence_factory, unflat_map,
                                 unflat_take, unflat_unique_rowid_map,
                                 unflat_vecmap, unflatten1, unflatten2, union,
                                 union_ordered, unique, unique_flags,
                                 unique_indices, unique_inverse,
                                 unique_ordered, unique_unordered,
                                 unpack_iterables, where, where_not_None,
                                 xor_lists, zipcompress, zipflat, ziptake,)
    from utool.util_num import (float_to_decimal, get_sys_maxfloat,
                                get_sys_maxint, get_sys_minint, int_comma_str,
                                num2_sigfig, num_fmt, order_of_magnitude_ceil,
                                sigfig_str,)
    from utool.util_numpy import (deterministic_sample, deterministic_shuffle,
                                  ensure_rng, index_of, intersect2d,
                                  make_incrementer, npfind, quantum_random,
                                  random_indexes, random_sample, sample_domain,
                                  shuffle, spaced_indexes, spaced_items,
                                  tiled_range,)
    from utool.util_path import (ChdirContext, IMG_EXTENSIONS, PRINT_CALLER,
                                 ancestor_paths, assert_exists, assertpath,
                                 augpath, basename_noext, checkpath, copy,
                                 copy_all, copy_files_to, copy_list,
                                 copy_single, delete, dirsplit,
                                 ensure_crossplat_path, ensure_ext,
                                 ensure_mingw_drive, ensure_native_path,
                                 ensure_unixslash, ensuredir, ensurefile,
                                 ensurepath, existing_commonprefix,
                                 existing_subpath, expand_win32_shortname,
                                 extend_regex, file_bytes, file_megabytes,
                                 find_lib_fpath, fnames_to_fpaths,
                                 fpath_has_ext, fpath_has_imgext,
                                 fpaths_to_fnames, get_modname_from_modpath,
                                 get_modpath, get_module_dir,
                                 get_module_subdir_list,
                                 get_nonconflicting_path, get_path_type,
                                 get_relative_modpath,
                                 get_standard_exclude_dnames,
                                 get_standard_include_patterns,
                                 get_win32_short_path_name, glob,
                                 glob_python_modules, grep, grepfile,
                                 greplines, iglob, is_module_dir,
                                 is_private_module, is_python_module, isfile,
                                 islink, ismount, list_images,
                                 longest_existing_path, ls, ls_dirs, ls_images,
                                 ls_moduledirs, ls_modulefiles,
                                 make_grep_resultstr, matching_fpaths, move,
                                 move_list, newcd, non_existing_path,
                                 num_images_in_dir, path_ndir_split,
                                 pathsplit_full, platform_path, relpath_unix,
                                 remove_broken_links, remove_dirs,
                                 remove_existing_fpaths, remove_file,
                                 remove_file_list, remove_files_in_dir,
                                 remove_fpaths, sanitize_filename,
                                 search_candidate_paths, search_in_dirs, sed,
                                 sedfile, splitdrive, symlink, tail, testgrep,
                                 touch, truepath, truepath_relative,
                                 unexpanduser, unixjoin, win_shortcut,)
    from utool.util_print import (Indenter, NO_INDENT, colorprint, cprint,
                                  dictprint, horiz_print, printNOTQUIET,
                                  printVERBOSE, printWARN, print_code,
                                  print_dict, print_difftext, print_filesize,
                                  print_list, print_locals, print_python_code,
                                  printdict, printif, printshape,)
    from utool.util_progress import (AGGROFLUSH, DEBUG_FREQ_ADJUST,
                                     FORCE_ALL_PROGRESS, NO_PROGRESS,
                                     PROGGRESS_BACKSPACE, ProgChunks, ProgIter,
                                     ProgPartial, ProgressIter,
                                     VALID_PROGRESS_TYPES, get_num_chunks,
                                     log_progress, progiter, progress_str,
                                     test_progress,)
    from utool.util_profile import (Profiler, cStringIO,
                                    clean_line_profile_text, clean_lprof_file,
                                    dump_profile_text, fix_rawprofile_blocks,
                                    get_block_id, get_block_totaltime,
                                    get_profile_text, get_summary,
                                    make_profiler, parse_rawprofile_blocks,
                                    parse_timemap_from_blocks,)
    from utool.util_project import (GrepResult, SetupRepo, UserProfile,
                                    ensure_text, ensure_user_profile,
                                    glob_projects, grep_projects,
                                    ibeis_user_profile, sed_projects,
                                    setup_repo,)
    from utool.util_parallel import (KillableProcess, KillableThread, bgfunc,
                                     buffered_generator, generate2,
                                     get_default_numprocs,
                                     get_sys_thread_limit, in_main_process,
                                     init_worker, set_num_procs,
                                     spawn_background_daemon_thread,
                                     spawn_background_process,
                                     spawn_background_thread,)
    from utool.util_resources import (available_memory, current_memory_usage,
                                      get_matching_process_ids,
                                      get_memstats_str,
                                      get_python_datastructure_sizes,
                                      get_resource_limits,
                                      get_resource_usage_str, memstats,
                                      num_cpus, num_unused_cpus, peak_memory,
                                      print_resource_usage, time_in_systemmode,
                                      time_in_usermode, time_str2,
                                      total_memory, used_memory,)
    from utool.util_str import (BACKSLASH, DOUBLE_QUOTE, ENABLE_COLORS,
                                NEWLINE, NO_TRUNCATE, SINGLE_QUOTE, TAUFMTSTR,
                                TAUSTR, TRIPLE_DOUBLE_QUOTE,
                                TRIPLE_SINGLE_QUOTE, align, align_lines,
                                array2string2, array_repr2, autoformat_pep8,
                                autopep8_format, bbox_str, bubbletext,
                                byte_str, byte_str2, chr_range, closet_words,
                                codeblock, color_diff_text, color_text,
                                conj_phrase, countdown_flag, dict_itemstr_list,
                                dict_str, difftext, doctest_code_line,
                                doctest_repr, ensure_ascii, ensure_unicode,
                                ensure_unicode_strlist, file_megabytes_str,
                                filesize_str, filtered_infostr, find_block_end,
                                flatten_textlines, format_multi_paragraphs,
                                format_multiple_paragraph_sentences,
                                format_single_paragraph_sentences,
                                format_text_as_docstr, func_callsig,
                                func_defsig, func_str, get_bytes,
                                get_callable_name, get_indentation,
                                get_itemstr_list, get_minimum_indentation,
                                get_textdiff, highlight_code,
                                highlight_multi_regex, highlight_regex,
                                highlight_text, horiz_string, hz_str, indent,
                                indent_list, indent_rest, indentjoin,
                                insert_block_between_lines,
                                is_byte_encoded_unicode, is_url, joins,
                                list_str, list_str_summarized,
                                long_fname_format, lorium_ipsum, msgblock,
                                number_text_lines, numpy_str,
                                order_of_magnitude_str, pack_into, packstr,
                                packtext, parse_bytes, pluralize, quantstr,
                                regex_reconstruct_split, remove_chars,
                                remove_doublenewlines, remove_doublspaces,
                                replace_between_tags, repr2, repr2_json, repr3,
                                repr4, reprfunc, scalar_str, second_str,
                                seconds_str, split_sentences2, str_between,
                                strip_ansi, testdata_text, textblock,
                                theta_str, to_camel_case, to_title_caps,
                                to_underscore_case, toggle_comment_lines,
                                trunc_repr, truncate_str,
                                unformat_text_as_docstr, unindent, utf8_len,
                                varinfo_str, verts_str,)
    from utool.util_sysreq import (ensure_in_pythonpath,
                                   get_global_dist_packages_dir,
                                   get_local_dist_packages_dir,
                                   get_site_packages_dir, in_virtual_env,
                                   is_running_as_root, locate_path,
                                   total_purge_developed_repo,)
    from utool.util_sqlite import (SQLColumnRichInfo,
                                   get_nonprimary_columninfo,
                                   get_primary_columninfo, get_table_column,
                                   get_table_columninfo_list,
                                   get_table_columnname_list,
                                   get_table_columns, get_table_csv,
                                   get_table_num_rows, get_table_rows,
                                   get_tablenames, print_database_structure,)
    from utool.util_setup import (NOOP, SETUP_PATTERNS, SetupManager,
                                  assert_in_setup_repo, autogen_sphinx_apidoc,
                                  build_pyo, clean, find_ext_modules,
                                  find_packages, get_cmdclass,
                                  get_numpy_include_dir, parse_author,
                                  parse_package_for_version, parse_readme,
                                  presetup, presetup_commands, read_license,
                                  setup_chmod, setuptools_setup,)
    from utool.util_set import (OrderedSet, oset,)
    from utool.util_regex import (REGEX_C_COMMENT, REGEX_ESCSTR, REGEX_FLOAT,
                                  REGEX_INT, REGEX_LATEX_COMMENT,
                                  REGEX_NONGREEDY, REGEX_RVAL, REGEX_STR,
                                  REGEX_VARNAME, REGEX_WHITESPACE, RE_FLAGS,
                                  RE_KWARGS, backref_field, bref_field,
                                  convert_text_to_varname, extend_regex2,
                                  extend_regex3, get_match_text,
                                  modify_quoted_strs, named_field,
                                  named_field_regex, named_field_repl,
                                  negative_lookahead, negative_lookbehind,
                                  nongreedy_kleene_star, padded_parse,
                                  parse_docblock, parse_python_syntax,
                                  positive_lookahead, positive_lookbehind,
                                  regex_get_match, regex_matches, regex_or,
                                  regex_parse, regex_replace,
                                  regex_replace_lines, regex_search,
                                  regex_split, regex_word, whole_word,)
    from utool.util_time import (Timer, Timerit, date_to_datetime,
                                 datetime_to_posixtime,
                                 determine_timestamp_format, ensure_timedelta,
                                 exiftime_to_unixtime, get_datestamp,
                                 get_posix_timedelta_str,
                                 get_posix_timedelta_str2, get_timedelta_str,
                                 get_timestamp, get_timestats_dict,
                                 get_timestats_str, get_unix_timedelta,
                                 get_unix_timedelta_str, local_timezone,
                                 parse_timedelta_str, parse_timestamp, tic,
                                 timestamp, toc, unixtime_to_datetimeobj,
                                 unixtime_to_datetimestr,
                                 unixtime_to_timedelta, utcnow_tz,)
    from utool.util_tags import (alias_tags, build_alias_map,
                                 filterflags_general_tags, modify_tags,
                                 tag_coocurrence, tag_hist,)
    from utool.util_type import (BooleanType, COMPARABLE_TYPES, FloatType,
                                 HAVE_PANDAS, IntType, LISTLIKE_TYPES,
                                 LongType, NUMPY_SCALAR_NAMES,
                                 NUMPY_TYPE_TUPLE, PRIMATIVE_TYPES,
                                 VALID_BOOL_TYPES, VALID_FLOAT_TYPES,
                                 VALID_INT_TYPES, VALID_STRING_TYPES,
                                 assert_int, bool_from_str, float_type_,
                                 fuzzy_int, fuzzy_subset,
                                 get_homogenous_list_type, get_type, int_type_,
                                 is_bool, is_comparable_type, is_dict,
                                 is_float, is_func_or_method,
                                 is_func_or_method_or_partial, is_funclike,
                                 is_int, is_list, is_listlike, is_method,
                                 is_str, is_tuple, is_type, is_valid_floattype,
                                 smart_cast, smart_cast2, try_cast, type_str,)
    from utool.util_tests import (BIGFACE, DEBUG_SRC, ExitTestException,
                                  HAPPY_FACE, HAPPY_FACE_BIG, HAPPY_FACE_SMALL,
                                  INDENT_TEST, ModuleDoctestTup, PRINT_FACE,
                                  PRINT_SRC, SAD_FACE, SAD_FACE_BIG,
                                  SAD_FACE_SMALL, SYSEXIT_ON_FAIL, TestTuple,
                                  VERBOSE_TEST, VERBOSE_TIMER, doctest_funcs,
                                  doctest_module_list, doctest_was_requested,
                                  execute_doctest, find_doctestable_modnames,
                                  find_testfunc, find_untested_modpaths,
                                  get_doctest_examples, get_module_completions,
                                  get_module_doctest_tup, get_module_testlines,
                                  get_package_testables, main_function_tester,
                                  parse_docblocks_from_docstr,
                                  parse_doctest_from_docstr, qt4ensure,
                                  qtensure, quit_if_noshow, read_exampleblock,
                                  run_test, show_if_requested,
                                  show_was_requested,)
    from utool.util_web import (find_open_port, get_localhost,
                                is_local_port_open, render_html,
                                start_simple_webserver,)
    from utool.DynamicStruct import (DynStruct,)
    from utool.Preferences import (Pref, PrefChoice, PrefInternal, PrefNode,
                                   PrefTree, VERBOSE_PREF, test_Preferences,)
    print, rrr, profile = util_inject.inject2(__name__, '[utool]')


    def reassign_submodule_attributes(verbose=1):
        """
        Updates attributes in the __init__ modules with updated attributes
        in the submodules.
        """
        import sys
        if verbose and '--quiet' not in sys.argv:
            print('dev reimport')
        # Self import
        import utool
        # Implicit reassignment.
        seen_ = set([])
        for tup in IMPORT_TUPLES:
            if len(tup) > 2 and tup[2]:
                continue  # dont import package names
            submodname, fromimports = tup[0:2]
            submod = getattr(utool, submodname)
            for attr in dir(submod):
                if attr.startswith('_'):
                    continue
                if attr in seen_:
                    # This just holds off bad behavior
                    # but it does mimic normal util_import behavior
                    # which is good
                    continue
                seen_.add(attr)
                setattr(utool, attr, getattr(submod, attr))


    def reload_subs(verbose=1):
        """ Reloads utool and submodules """
        if verbose:
            print('Reloading utool submodules')
        rrr(verbose > 1)
        def wrap_fbrrr(mod):
            def fbrrr(*args, **kwargs):
                """ fallback reload """
                if verbose > 0:
                    print('Auto-reload (using rrr) not setup for mod=%r' % (mod,))
            return fbrrr
        def get_rrr(mod):
            if hasattr(mod, 'rrr'):
                return mod.rrr
            else:
                return wrap_fbrrr(mod)
        def get_reload_subs(mod):
            return getattr(mod, 'reload_subs', wrap_fbrrr(mod))
        get_rrr(_internal)(verbose > 1)
        get_rrr(util_alg)(verbose > 1)
        get_rrr(util_aliases)(verbose > 1)
        get_rrr(util_arg)(verbose > 1)
        get_rrr(util_assert)(verbose > 1)
        get_rrr(util_autogen)(verbose > 1)
        get_rrr(util_cache)(verbose > 1)
        get_rrr(util_cplat)(verbose > 1)
        get_rrr(util_class)(verbose > 1)
        get_rrr(util_const)(verbose > 1)
        get_rrr(util_csv)(verbose > 1)
        get_rrr(util_config)(verbose > 1)
        get_rrr(util_dbg)(verbose > 1)
        get_rrr(util_dev)(verbose > 1)
        get_rrr(util_decor)(verbose > 1)
        get_rrr(util_dict)(verbose > 1)
        get_rrr(util_func)(verbose > 1)
        get_rrr(util_grabdata)(verbose > 1)
        get_rrr(util_gridsearch)(verbose > 1)
        get_rrr(util_git)(verbose > 1)
        get_rrr(util_latex)(verbose > 1)
        get_rrr(util_graph)(verbose > 1)
        get_rrr(util_hash)(verbose > 1)
        get_rrr(util_import)(verbose > 1)
        get_rrr(util_inject)(verbose > 1)
        get_rrr(util_io)(verbose > 1)
        get_rrr(util_iter)(verbose > 1)
        get_rrr(util_inspect)(verbose > 1)
        get_rrr(util_ipynb)(verbose > 1)
        get_rrr(util_logging)(verbose > 1)
        get_rrr(util_list)(verbose > 1)
        get_rrr(util_num)(verbose > 1)
        get_rrr(util_numpy)(verbose > 1)
        get_rrr(util_path)(verbose > 1)
        get_rrr(util_print)(verbose > 1)
        get_rrr(util_progress)(verbose > 1)
        get_rrr(util_profile)(verbose > 1)
        get_rrr(util_project)(verbose > 1)
        get_rrr(util_parallel)(verbose > 1)
        get_rrr(util_resources)(verbose > 1)
        get_rrr(util_str)(verbose > 1)
        get_rrr(util_sysreq)(verbose > 1)
        get_rrr(util_sqlite)(verbose > 1)
        get_rrr(util_setup)(verbose > 1)
        get_rrr(util_set)(verbose > 1)
        get_rrr(util_regex)(verbose > 1)
        get_rrr(util_time)(verbose > 1)
        get_rrr(util_tags)(verbose > 1)
        get_rrr(util_type)(verbose > 1)
        get_rrr(util_tests)(verbose > 1)
        get_rrr(util_web)(verbose > 1)
        get_rrr(DynamicStruct)(verbose > 1)
        get_rrr(Preferences)(verbose > 1)
        rrr(verbose > 1)
        try:
            # hackish way of propogating up the new reloaded submodule attributes
            reassign_submodule_attributes(verbose=verbose)
        except Exception as ex:
            print(ex)
    rrrr = reload_subs
    # </AUTOGEN_INIT>
